import copy

from django.utils.datastructures import MultiValueDict

from djblets.conditions.choices import (BaseConditionChoice,
                                        BaseConditionIntegerChoice,
                                        ConditionChoices)
from djblets.conditions.operators import (BaseConditionOperator,
                                          ConditionOperators)
from djblets.conditions.values import (ConditionValueCharField,
                                       ConditionValueIntegerField)
from djblets.forms.fields import ConditionsField
from djblets.testing.testcases import TestCase


class ConditionsWidgetTests(TestCase):
    """Unit tests for djblets.forms.widgets.ConditionsWidget."""

    def test_deepcopy(self):
        """Testing ConditionsWidget.__deepcopy__"""
        class MyChoice(BaseConditionIntegerChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)
        widget1 = field.widget
        widget2 = copy.deepcopy(widget1)

        widget1.mode_widget.attrs['foo'] = True
        widget1.choice_widget.attrs['foo'] = True
        widget1.operator_widget.attrs['foo'] = True
        widget1.condition_errors[0] = 'This is a test.'

        self.assertEqual(widget2.mode_widget.attrs, {})
        self.assertEqual(widget2.choice_widget.attrs, {})
        self.assertEqual(widget2.operator_widget.attrs, {})
        self.assertEqual(widget2.condition_errors, {})

        # Choices won't be modified between copies, and is a shared object,
        # so both should have the same instance.
        self.assertIs(widget1.choices, widget2.choices)

    def test_value_from_datadict(self):
        """Testing ConditionsWidget.value_from_datadict"""
        class MyChoice(BaseConditionIntegerChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        data = MultiValueDict('')
        data.update({
            'my_conditions_mode': 'any',
            'my_conditions_last_id': '1',
            'my_conditions_choice[0]': 'my-choice',
            'my_conditions_operator[0]': 'is',
            'my_conditions_value[0]': 'my-value-1',
            'my_conditions_choice[1]': 'my-choice',
            'my_conditions_operator[1]': 'is-not',
            'my_conditions_value[1]': 'my-value-2',
        })

        self.assertEqual(
            field.widget.value_from_datadict(data, MultiValueDict(''),
                                             'my_conditions'),
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'my-choice',
                        'op': 'is',
                        'value': 'my-value-1',
                    },
                    {
                        'choice': 'my-choice',
                        'op': 'is-not',
                        'value': 'my-value-2',
                    },
                ],
            })

    def test_value_from_datadict_with_missing_data(self):
        """Testing ConditionsWidget.value_from_datadict with missing data"""
        class MyChoice(BaseConditionIntegerChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        data = MultiValueDict('')

        self.assertEqual(
            field.widget.value_from_datadict(data, MultiValueDict(''),
                                             'my_conditions'),
            {
                'mode': None,
                'conditions': [],
            })

    def test_value_from_datadict_with_missing_last_id(self):
        """Testing ConditionsWidget.value_from_datadict with missing last_id"""
        class MyChoice(BaseConditionIntegerChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        data = MultiValueDict('')
        data.update({
            'my_conditions_mode': 'any',
            'my_conditions_choice[0]': 'my-choice',
            'my_conditions_operator[0]': 'is',
            'my_conditions_value[0]': 'my-value-1',
            'my_conditions_choice[1]': 'my-choice',
            'my_conditions_operator[1]': 'is-not',
            'my_conditions_value[1]': 'my-value-2',
        })

        self.assertEqual(
            field.widget.value_from_datadict(data, MultiValueDict(''),
                                             'my_conditions'),
            {
                'mode': 'any',
                'conditions': [],
            })

    def test_value_from_datadict_with_missing_choice_rows(self):
        """Testing ConditionsWidget.value_from_datadict with missing choice
        rows
        """
        class MyChoice(BaseConditionIntegerChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        data = MultiValueDict('')
        data.update({
            'my_conditions_mode': 'any',
            'my_conditions_last_id': '5',
            'my_conditions_choice[5]': 'my-choice',
            'my_conditions_operator[5]': 'is-not',
            'my_conditions_value[5]': 'my-value',
        })

        self.assertEqual(
            field.widget.value_from_datadict(data, MultiValueDict(''),
                                             'my_conditions'),
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'my-choice',
                        'op': 'is-not',
                        'value': 'my-value',
                    },
                ],
            })

    def test_get_context(self):
        """Testing ConditionsWidget.get_context"""
        class MyOperator1(BaseConditionOperator):
            operator_id = 'my-op-1'
            name = 'My Op 1'
            value_field = ConditionValueIntegerField()

        class MyOperator2(BaseConditionOperator):
            operator_id = 'my-op-2'
            name = 'My Op 2'
            value_field = ConditionValueCharField()

        class MyOperator3(BaseConditionOperator):
            operator_id = 'my-op-3'
            name = 'My Op 3'

        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice-1'
            name = 'My Choice 1'
            operators = ConditionOperators([MyOperator1, MyOperator2])

        class MyChoice2(BaseConditionChoice):
            choice_id = 'my-choice-2'
            name = 'My Choice 2'
            operators = ConditionOperators([MyOperator3])

        choices = ConditionChoices([MyChoice1, MyChoice2])
        field = ConditionsField(choices=choices)

        result = field.widget.get_context(
            'my_conditions',
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'my-choice-1',
                        'op': 'my-op-1',
                        'value': 'my-value-1',
                    },
                    {
                        'choice': 'my-choice-2',
                        'op': 'my-op-3',
                    },
                ],
            },
            {
                'id': 'my-conditions',
            })

        self.assertEqual(
            set(result),
            {
                'field_id', 'field_name', 'rendered_mode', 'rendered_rows',
                'serialized_choices', 'serialized_rows',
            })
        self.assertEqual(result['field_id'], 'my-conditions')
        self.assertEqual(result['field_name'], 'my_conditions')
        self.assertHTMLEqual(
            result['rendered_mode'],
            '<ul id="my_conditions_mode">\n'
            '<li><label for="my_conditions_mode_0">'
            '<input id="my_conditions_mode_0" name="my_conditions_mode"'
            ' type="radio" value="always" />'
            ' Always match</label></li>\n'
            '<li><label for="my_conditions_mode_1">'
            '<input id="my_conditions_mode_1" name="my_conditions_mode"'
            ' type="radio" value="all" />'
            ' Match <b>all</b> of the following:</label></li>\n'
            '<li><label for="my_conditions_mode_2">'
            '<input checked="checked" id="my_conditions_mode_2" '
            'name="my_conditions_mode" type="radio" value="any" />'
            ' Match <b>any</b> of the following:</label></li>\n'
            '</ul>')

        rendered_rows = result['rendered_rows']
        self.assertEqual(len(rendered_rows), 2)

        # Check the first rendered row.
        rendered_row = rendered_rows[0]
        self.assertEqual(set(rendered_row.keys()),
                         {'choice', 'operator', 'error'})

        self.assertHTMLEqual(
            rendered_row['choice'],
            '<select id="my-conditions_choice_0"'
            ' name="my_conditions_choice[0]">\n'
            '<option value="my-choice-1" selected="selected">'
            'My Choice 1</option>\n'
            '<option value="my-choice-2">'
            'My Choice 2</option>\n'
            '</select>')

        self.assertHTMLEqual(
            rendered_row['operator'],
            '<select id="my-conditions_operator_0"'
            ' name="my_conditions_operator[0]">\n'
            '<option value="my-op-1" selected="selected">'
            'My Op 1</option>\n'
            '<option value="my-op-2">My Op 2</option>\n'
            '</select>')

        self.assertIsNone(rendered_row['error'])

        # Check the second rendered row.
        rendered_row = rendered_rows[1]
        self.assertEqual(set(rendered_row.keys()),
                         {'choice', 'operator', 'error'})

        self.assertHTMLEqual(
            rendered_row['choice'],
            '<select id="my-conditions_choice_1"'
            ' name="my_conditions_choice[1]">\n'
            '<option value="my-choice-1">'
            'My Choice 1</option>\n'
            '<option value="my-choice-2" selected="selected">'
            'My Choice 2</option>\n'
            '</select>')

        self.assertHTMLEqual(
            rendered_row['operator'],
            '<select id="my-conditions_operator_1"'
            ' name="my_conditions_operator[1]">\n'
            '<option value="my-op-3" selected="selected">'
            'My Op 3</option>\n'
            '</select>')

        self.assertIsNone(rendered_row['error'])

        serialized_choices = result['serialized_choices']
        self.assertEqual(len(serialized_choices), 2)

        # Check the first serialized choice.
        serialized_choice = serialized_choices[0]
        self.assertEqual(set(serialized_choice.keys()),
                         {'id', 'name', 'valueField', 'operators'})
        self.assertEqual(serialized_choice['id'], 'my-choice-1')
        self.assertEqual(serialized_choice['name'], 'My Choice 1')
        self.assertEqual(serialized_choice['valueField'], {})
        self.assertEqual(serialized_choice['valueField'], {})

        serialized_operators = serialized_choice['operators']
        self.assertEqual(len(serialized_operators), 2)

        # Check the first operator in the first serialized choice.
        serialized_operator = serialized_operators[0]
        self.assertEqual(set(serialized_operator.keys()),
                         {'id', 'name', 'useValue', 'valueField'})
        self.assertEqual(serialized_operator['id'], 'my-op-1')
        self.assertEqual(serialized_operator['name'], 'My Op 1')
        self.assertTrue(serialized_operator['useValue'])

        serialized_value_field = serialized_operator['valueField']
        self.assertEqual(set(serialized_value_field.keys()),
                         {'model', 'view'})

        serialized_value_model = serialized_value_field['model']
        self.assertEqual(serialized_value_model['className'],
                         'Djblets.Forms.ConditionValueField')
        self.assertHTMLEqual(serialized_value_model['data']['fieldHTML'],
                             '<input name="XXX" type="number" />')

        serialized_value_view = serialized_value_field['view']
        self.assertEqual(serialized_value_view['className'],
                         'Djblets.Forms.ConditionValueFormFieldView')
        self.assertEqual(serialized_value_view['data'], {})

        # Check the second operator in the first serialized choice.
        serialized_operator = serialized_operators[1]
        self.assertEqual(set(serialized_operator.keys()),
                         {'id', 'name', 'useValue', 'valueField'})
        self.assertEqual(serialized_operator['id'], 'my-op-2')
        self.assertEqual(serialized_operator['name'], 'My Op 2')
        self.assertTrue(serialized_operator['useValue'])

        serialized_value_field = serialized_operator['valueField']
        self.assertEqual(set(serialized_value_field.keys()),
                         {'model', 'view'})

        serialized_value_model = serialized_value_field['model']
        self.assertEqual(serialized_value_model['className'],
                         'Djblets.Forms.ConditionValueField')
        self.assertHTMLEqual(serialized_value_model['data']['fieldHTML'],
                             '<input name="XXX" type="text" />')

        serialized_value_view = serialized_value_field['view']
        self.assertEqual(serialized_value_view['className'],
                         'Djblets.Forms.ConditionValueFormFieldView')
        self.assertEqual(serialized_value_view['data'], {})

        self.assertEqual(
            serialized_choices[1],
            {
                'id': 'my-choice-2',
                'name': 'My Choice 2',
                'operators': [
                    {
                        'id': 'my-op-3',
                        'name': 'My Op 3',
                        'useValue': False,
                    },
                ],
                'valueField': {},
            })

        serialized_rows = result['serialized_rows']
        self.assertEqual(len(serialized_rows), 2)
        self.assertEqual(
            serialized_rows[0],
            {
                'choiceID': 'my-choice-1',
                'operatorID': 'my-op-1',
                'valid': True,
                'value': 'my-value-1',
            })
        self.assertEqual(
            serialized_rows[1],
            {
                'choiceID': 'my-choice-2',
                'operatorID': 'my-op-3',
                'valid': True,
                'value': None,
            })

    def test_get_context_with_invalid_choice(self):
        """Testing ConditionsWidget.get_context with invalid choice"""
        class MyOperator1(BaseConditionOperator):
            operator_id = 'my-op-1'
            name = 'My Op 1'
            value_field = ConditionValueIntegerField()

        class MyOperator2(BaseConditionOperator):
            operator_id = 'my-op-2'
            name = 'My Op 2'
            value_field = ConditionValueCharField()

        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice-1'
            name = 'My Choice 1'
            operators = ConditionOperators([MyOperator1, MyOperator2])

        choices = ConditionChoices([MyChoice1])
        field = ConditionsField(choices=choices)

        result = field.widget.get_context(
            'my_conditions',
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'invalid-choice',
                        'op': 'my-op-1',
                        'value': 'my-value-1',
                    },
                ],
            },
            {
                'id': 'my-conditions',
            })

        rendered_rows = result['rendered_rows']
        self.assertEqual(len(rendered_rows), 1)

        rendered_row = rendered_rows[0]
        self.assertEqual(set(rendered_row.keys()),
                         {'choice', 'operator', 'error'})
        self.assertHTMLEqual(
            rendered_row['choice'],
            '<select disabled="disabled" id="my-conditions_choice_0"'
            ' name="my_conditions_choice[0]">\n'
            '<option value="my-choice-1">My Choice 1</option>\n'
            '<option value="invalid-choice" selected="selected">'
            'invalid-choice</option>\n'
            '</select>'
            '<input name="my_conditions_choice[0]" type="hidden"'
            ' value="invalid-choice" />')

        self.assertHTMLEqual(
            rendered_row['operator'],
            '<select disabled="disabled" id="my-conditions_operator_0"'
            ' name="my_conditions_operator[0]">\n'
            '<option value="my-op-1" selected="selected">'
            'my-op-1</option>\n'
            '</select>'
            '<input name="my_conditions_operator[0]" type="hidden"'
            ' value="my-op-1" />')

        self.assertEqual(
            rendered_row['error'],
            'This choice no longer exists. You will need to '
            'delete the condition in order to make changes.')

        serialized_choices = result['serialized_choices']
        self.assertEqual(len(serialized_choices), 1)
        self.assertEqual(serialized_choices[0]['id'], 'my-choice-1')

        serialized_rows = result['serialized_rows']
        self.assertEqual(
            serialized_rows,
            [{
                'choiceID': 'invalid-choice',
                'operatorID': 'my-op-1',
                'valid': False,
                'value': 'my-value-1',
                'error': ('This choice no longer exists. You will need to '
                          'delete the condition in order to make changes.'),
            }])

    def test_get_context_with_invalid_operator(self):
        """Testing ConditionsWidget.get_context with invalid operator"""
        class MyOperator1(BaseConditionOperator):
            operator_id = 'my-op-1'
            name = 'My Op 1'
            value_field = ConditionValueIntegerField()

        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice'
            name = 'My Choice'
            operators = ConditionOperators([MyOperator1])

        choices = ConditionChoices([MyChoice1])
        field = ConditionsField(choices=choices)

        result = field.widget.get_context(
            'my_conditions',
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'my-choice',
                        'op': 'invalid-op',
                        'value': 'my-value-1',
                    },
                ],
            },
            {
                'id': 'my-conditions',
            })

        rendered_rows = result['rendered_rows']
        self.assertEqual(len(rendered_rows), 1)

        rendered_row = rendered_rows[0]
        self.assertEqual(set(rendered_row.keys()),
                         {'choice', 'operator', 'error'})
        self.assertHTMLEqual(
            rendered_row['choice'],
            '<select disabled="disabled" id="my-conditions_choice_0"'
            ' name="my_conditions_choice[0]">\n'
            '<option value="my-choice" selected="selected">'
            'My Choice</option>\n'
            '</select>'
            '<input name="my_conditions_choice[0]" type="hidden"'
            ' value="my-choice" />')

        self.assertHTMLEqual(
            rendered_row['operator'],
            '<select disabled="disabled" id="my-conditions_operator_0"'
            ' name="my_conditions_operator[0]">\n'
            '<option value="invalid-op" selected="selected">'
            'invalid-op</option>\n'
            '</select>'
            '<input name="my_conditions_operator[0]" type="hidden"'
            ' value="invalid-op" />')

        self.assertEqual(
            rendered_row['error'],
            'This operator no longer exists. You will need to '
            'delete the condition in order to make changes.')

        serialized_choices = result['serialized_choices']
        self.assertEqual(len(serialized_choices), 1)
        self.assertEqual(serialized_choices[0]['id'], 'my-choice')

        serialized_rows = result['serialized_rows']
        self.assertEqual(
            serialized_rows,
            [{
                'choiceID': 'my-choice',
                'operatorID': 'invalid-op',
                'valid': False,
                'value': 'my-value-1',
                'error': ('This operator no longer exists. You will need to '
                          'delete the condition in order to make changes.'),
            }])

    def test_get_context_with_condition_errors(self):
        """Testing ConditionsWidget.get_context with condition errors"""
        class MyOperator1(BaseConditionOperator):
            operator_id = 'my-op-1'
            name = 'My Op 1'
            value_field = ConditionValueIntegerField()

        class MyChoice1(BaseConditionChoice):
            choice_id = 'my-choice'
            name = 'My Choice'
            operators = ConditionOperators([MyOperator1])

        choices = ConditionChoices([MyChoice1])
        field = ConditionsField(choices=choices)
        field.widget.condition_errors[0] = 'This is an error.'

        result = field.widget.get_context(
            'my_conditions',
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'my-choice',
                        'op': 'my-op-1',
                        'value': 'my-value-1',
                    },
                ],
            },
            {
                'id': 'my-conditions',
            })

        rendered_rows = result['rendered_rows']
        self.assertEqual(len(rendered_rows), 1)

        rendered_row = rendered_rows[0]
        self.assertEqual(set(rendered_row.keys()),
                         {'choice', 'operator', 'error'})
        self.assertHTMLEqual(
            rendered_row['choice'],
            '<select id="my-conditions_choice_0"'
            ' name="my_conditions_choice[0]">\n'
            '<option value="my-choice" selected="selected">'
            'My Choice</option>\n'
            '</select>')

        self.assertHTMLEqual(
            rendered_row['operator'],
            '<select id="my-conditions_operator_0"'
            ' name="my_conditions_operator[0]">\n'
            '<option value="my-op-1" selected="selected">'
            'My Op 1</option>\n'
            '</select>')

        self.assertEqual(rendered_row['error'], 'This is an error.')

        serialized_choices = result['serialized_choices']
        self.assertEqual(len(serialized_choices), 1)
        self.assertEqual(serialized_choices[0]['id'], 'my-choice')

        serialized_rows = result['serialized_rows']
        self.assertEqual(
            serialized_rows,
            [{
                'choiceID': 'my-choice',
                'operatorID': 'my-op-1',
                'valid': True,
                'value': 'my-value-1',
                'error': 'This is an error.',
            }])

    def test_render(self):
        """Testing ConditionsWidget.render"""
        class MyOperator(BaseConditionOperator):
            operator_id = 'my-op'
            name = 'My Op'
            value_field = ConditionValueIntegerField()

        class MyChoice(BaseConditionChoice):
            choice_id = 'my-choice'
            name = 'My Choice'
            operators = ConditionOperators([MyOperator])

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        rendered = field.widget.render(
            'my_conditions',
            {
                'mode': 'any',
                'conditions': [
                    {
                        'choice': 'my-choice',
                        'op': 'my-op',
                        'value': 'my-value-1',
                    },
                ],
            },
            {
                'id': 'my-conditions',
            })

        self.assertIn('<div class="conditions-field" id="my-conditions">',
                      rendered)
        self.assertIn('<input type="hidden" name="my_conditions_last_id"'
                      ' value="1">',
                      rendered)
        self.assertRegex(
            rendered,
            '<option value="my-choice" selected(="selected")?>My Choice'
            '</option>')
        self.assertRegex(
            rendered,
            '<option value="my-op" selected(="selected")?>My Op</option>')
        self.assertIn('<span class="conditions-field-value"></span>',
                      rendered)



"""Tests for djblets.extensions.templatetags."""

import os
import re
from collections import OrderedDict

from django.http import HttpRequest
from django.template import Context, Template
from django.urls import ResolverMatch
from django.utils.safestring import mark_safe
from kgb import SpyAgency

from djblets.extensions.extension import Extension, ExtensionInfo, JSExtension
from djblets.extensions.hooks import TemplateHook
from djblets.extensions.manager import ExtensionManager
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.testing.testcases import TestCase
from pipeline.conf import settings as pipeline_settings


class MyTestJSExtension1(JSExtension):
    model_class = 'TestExtension1'

    def get_model_data(self, request, **kwargs):
        return {'test': 'new'}


class MyTestJSExtension2(JSExtension):
    model_class = 'TestExtension2'

    def get_model_data(self, **kwargs):
        return {'test': 'old'}


class MyTestExtensionManager(ExtensionManager):
    should_install_static_media = True


class MyTestExtension(Extension):
    # While normally these would be plain dictionaries, we want to ensure
    # order for testing, so we'll be doing this a bit differently.
    css_bundles = OrderedDict()
    css_bundles['default'] = {
        'source_filenames': (
            'css/default-test.css',
        ),
    }
    css_bundles['optional'] = {
        'source_filenames': (
            'css/optional-test.css',
        ),
        'apply_to': ['foo'],
    }

    js_bundles = OrderedDict()
    js_bundles['default'] = {
        'source_filenames': (
            'js/default-test.js',
        ),
    }
    js_bundles['optional'] = {
        'source_filenames': (
            'js/optional-test.js',
        ),
        'apply_to': ['foo'],
    }

    js_extensions = [MyTestJSExtension1, MyTestJSExtension2]


class TemplateTagTests(SpyAgency, ExtensionTestCaseMixin, TestCase):
    """Tests for djblets.extensions.templatetags."""

    default_extension_manager_cls = MyTestExtensionManager

    def setUp(self):
        def _has_resource(self, path):
            return path == 'static'

        def _extract_resource(self, path):
            if path == 'static':
                return os.path.abspath(os.path.join(__file__, '..', 'static'))

            return None

        super(TemplateTagTests, self).setUp()

        self.spy_on(ExtensionInfo.has_resource,
                    owner=ExtensionInfo,
                    call_fake=_has_resource)
        self.spy_on(ExtensionInfo.extract_resource,
                    owner=ExtensionInfo,
                    call_fake=_extract_resource)

        self.extension = self.setup_extension(MyTestExtension)
        self.request = HttpRequest()

        self._old_pipeline_enabled = pipeline_settings.PIPELINE_ENABLED

    def tearDown(self):
        super(TemplateTagTests, self).tearDown()

        pipeline_settings.PIPELINE_ENABLED = self._old_pipeline_enabled

    def test_template_hook_point_tag(self):
        """Testing template_hook_point template tag"""
        class DummyTemplateHook(TemplateHook):
            """A template hook where the content is specified."""

            def __init__(self, content, *args, **kwargs):
                super(DummyTemplateHook, self).__init__(*args, **kwargs)
                self._content = content

            def render_to_string(self, request, context):
                return self._content

        t = Template('{% load djblets_extensions %}'
                     '{% template_hook_point "test-hook" %}')

        DummyTemplateHook(content=mark_safe('<strong>foo</strong>\n'),
                          extension=self.extension,
                          name='test-hook')
        DummyTemplateHook(content='<bar>\n',
                          extension=self.extension,
                          name='test-hook')

        self.assertEqual(
            t.render(Context({
                'request': self.request,
            })),
            '<strong>foo</strong>\n'
            '&lt;bar&gt;\n')

    def test_ext_static_tag(self):
        """Testing ext_static template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% ext_static ext "css/default-test.css" %}')

        self.assertEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '/ext/djblets.extensions.tests.test_templatetags.MyTestExtension/'
            'css/default-test.dad0c9b31e59.css')

    def test_ext_css_bundle_tag_with_pipeline_enabled(self):
        """Testing ext_css_bundle template tag with PIPELINE_ENABLED=True"""
        pipeline_settings.PIPELINE_ENABLED = True

        t = Template('{% load djblets_extensions %}'
                     '{% ext_css_bundle ext "default" %}')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/css/default.min.dad0c9b31e59.css" '
            'media="all" rel="stylesheet" type="text/css" />')

    def test_ext_css_bundle_tag_with_pipeline_disabled(self):
        """Testing ext_css_bundle template tag with PIPELINE_ENABLED=False"""
        pipeline_settings.PIPELINE_ENABLED = False

        t = Template('{% load djblets_extensions %}'
                     '{% ext_css_bundle ext "default" %}')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '<link href="/ext/djblets.extensions.tests.test_templatetags.'
            'MyTestExtension/css/default-test.dad0c9b31e59.css" '
            'media="all" rel="stylesheet" type="text/css">')

    def test_ext_js_bundle_tag_with_pipeline_enabled(self):
        """Testing ext_js_bundle template tag with PIPELINE_ENABLED=True"""
        pipeline_settings.PIPELINE_ENABLED = True

        t = Template('{% load djblets_extensions %}'
                     '{% ext_js_bundle ext "default" %}')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetag'
            's.MyTestExtension/js/default.min.dad0c9b31e59.js" '
            'charset="utf-8"></script>')

    def test_ext_js_bundle_tag_with_pipeline_disabled(self):
        """Testing ext_js_bundle template tag with PIPELINE_ENABLED=False"""
        pipeline_settings.PIPELINE_ENABLED = False

        t = Template('{% load djblets_extensions %}'
                     '{% ext_js_bundle ext "default" %}')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetag'
            's.MyTestExtension/js/default-test.dad0c9b31e59.js" '
            'charset="utf-8"></script>')

    def test_load_extensions_css_tag_with_pipline_enabled(self):
        """Testing load_extensions_css template tag with PIPELINE_ENABLED=True
        """
        pipeline_settings.PIPELINE_ENABLED = True

        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_css manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.extension_mgr.key,
                'request': self.request,
            })),
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/css/default.min.dad0c9b31e59.css" '
            'rel="stylesheet" media="all" type="text/css">'
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/css/optional.min.dad0c9b31e59.css" '
            'rel="stylesheet" media="all" type="text/css">')

    def test_load_extensions_css_tag_with_pipline_disabled(self):
        """Testing load_extensions_css template tag with PIPELINE_ENABLED=False
        """
        pipeline_settings.PIPELINE_ENABLED = False

        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_css manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.extension_mgr.key,
                'request': self.request,
            })),
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/css/default-test.dad0c9b31e59.css" '
            'rel="stylesheet" media="all" type="text/css">'
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/css/optional-test.dad0c9b31e59.css" '
            'rel="stylesheet" media="all" type="text/css">')

    def test_load_extensions_js_tag_with_pipeline_enabled(self):
        """Testing load_extensions_js template tag with PIPELINE_ENABLED=True
        """
        pipeline_settings.PIPELINE_ENABLED = True

        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_js manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.extension_mgr.key,
                'request': self.request,
            })),
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/js/default.min.dad0c9b31e59.js" '
            'charset="utf-8"></script>'
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/js/optional.min.dad0c9b31e59.js" '
            'charset="utf-8"></script>')

    def test_load_extensions_js_tag_with_pipeline_disabled(self):
        """Testing load_extensions_js template tag with PIPELINE_ENABLED=False
        """
        pipeline_settings.PIPELINE_ENABLED = False

        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_js manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.extension_mgr.key,
                'request': self.request,
            })),
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/js/default-test.dad0c9b31e59.js" '
            'charset="utf-8"></script>'
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetags'
            '.MyTestExtension/js/optional-test.dad0c9b31e59.js" '
            'charset="utf-8"></script>')

    def test_init_js_extensions(self):
        """Testing init_js_extensions template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% init_js_extensions manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        content = t.render(Context({
            'ext': self.extension,
            'manager_id': self.extension_mgr.key,
            'request': self.request,
        }))

        self.assertIsNotNone(re.search(
            r'new TestExtension1\({\s+"test": "new",',
            content))

        self.assertIsNotNone(re.search(
            r'new TestExtension2\({\s+"test": "old",',
            content))



"""Unit tests for djblets.privacy.consent.forms.ConsentRequirementField."""

from django.contrib.auth.models import User
from django.core.exceptions import ValidationError

from djblets.privacy.consent import (BaseConsentRequirement, Consent,
                                     get_consent_tracker)
from djblets.privacy.consent.forms import ConsentRequirementField
from djblets.privacy.tests.testcases import ConsentTestCase


class MyConsentRequirement(BaseConsentRequirement):
    requirement_id = 'my-requirement'
    name = 'My Requirement'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'
    icons = {
        '1x': '/static/consent.png',
        '2x': '/static/consent@2x.png',
    }


class ConsentRequirementFieldTests(ConsentTestCase):
    """Unit tests for ConsentRequirementField."""

    def setUp(self):
        super(ConsentRequirementFieldTests, self).setUp()

        self.consent_requirement = MyConsentRequirement()

        self.field = ConsentRequirementField(
            consent_requirement=self.consent_requirement,
            consent_source='https://example.com/consent/',
            extra_consent_data={
                'test': True,
            })

        self.user = User.objects.create(username='test-user')

    def test_init_with_user_and_no_existing_consent(self):
        """Testing ConsentRequirementField.__init__ with user and no existing
        consent data
        """
        self.field.set_initial_from_user(self.user)
        self.assertEqual(self.field.initial, Consent.UNSET)

    def test_init_with_user_and_existing_consent(self):
        """Testing ConsentRequirementField.__init__ with user and existing
        consent data
        """
        get_consent_tracker().record_consent_data(
            self.user,
            self.consent_requirement.build_consent_data(granted=True))

        self.field.set_initial_from_user(self.user)
        self.assertEqual(self.field.initial, Consent.GRANTED)

    def test_set_initial_from_user(self):
        """Testing ConsentRequirementField.set_initial_from_user"""
        get_consent_tracker().record_consent_data(
            self.user,
            self.consent_requirement.build_consent_data(granted=False))

        self.field.set_initial_from_user(self.user)
        self.assertEqual(self.field.initial, Consent.DENIED)

    def test_prepare_value(self):
        """Testing ConsentRequirementField.prepare_value"""
        self.assertEqual(self.field.prepare_value(Consent.GRANTED), 'allow')
        self.assertEqual(self.field.prepare_value(Consent.DENIED), 'block')
        self.assertIsNone(self.field.prepare_value(Consent.UNSET))

    def test_clean_with_allow(self):
        """Testing ConsentRequirementField.clean with allow"""
        consent_data = self.field.clean('allow')

        self.assertIsNotNone(consent_data)
        self.assertEqual(consent_data.requirement_id, 'my-requirement')
        self.assertEqual(consent_data.source, 'https://example.com/consent/')
        self.assertTrue(consent_data.granted)
        self.assertEqual(
            consent_data.extra_data,
            {
                'test': True,
            })

    def test_clean_with_block(self):
        """Testing ConsentRequirementField.clean with block"""
        consent_data = self.field.clean('block')

        self.assertIsNotNone(consent_data)
        self.assertEqual(consent_data.requirement_id, 'my-requirement')
        self.assertEqual(consent_data.source, 'https://example.com/consent/')
        self.assertFalse(consent_data.granted)
        self.assertEqual(
            consent_data.extra_data,
            {
                'test': True,
            })

    def test_clean_with_unset(self):
        """Testing ConsentRequirementField.clean with unset"""
        message = 'You must choose Allow or Block to continue.'

        with self.assertRaisesMessage(ValidationError, message):
            self.field.clean(None)



"""Utilities and constants for configuring Pipeline.

Note:
    This is safe to import from a project's :file:`settings.py` without
    side effects (for instance, it will not load any Django models or apps).

Version Added:
    2.1
"""

import os

from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import force_str


#: Default list of compilers used by Djblets.
DEFAULT_PIPELINE_COMPILERS = [
    'djblets.pipeline.compilers.es6.ES6Compiler',
    'djblets.pipeline.compilers.less.LessCompiler',
]


def build_pipeline_settings(pipeline_enabled,
                            node_modules_path,
                            static_root,
                            javascript_bundles=[],
                            stylesheet_bundles=[],
                            compilers=DEFAULT_PIPELINE_COMPILERS,
                            babel_extra_plugins=[],
                            babel_extra_args=[],
                            less_extra_args=[],
                            validate_paths=True):
    """Build a standard set of Pipeline settings.

    This can be used to create a ``PIPELINE`` settings dictionary in a
    :file:`settings.py` file based on the standard Djblets Pipeline settings,
    which makes use of Babel, LessCSS, and UglifyJS, along with a preset
    list of plugins.

    The following base set of Babel plugins are used:

    * `dedent <https://www.npmjs.com/package/babel-plugin-dedent>`_
    * `django-gettext
      <https://www.npmjs.com/package/babel-plugin-django-gettext>`_

    The following LessCSS plugin is used:

    * `autoprefix
      <https://www.npmjs.com/package/@beanbag/less-plugin-autoprefix>`_

    This will also set the value of ``node_modules_path`` in the
    :envvar:`NODE_PATH` environment variable.

    Args:
        pipeline_enabled (bool):
            Whether Pipelining of static media should be enabled.

            This must be provided by a caller. It's recommended to enable
            this if ``DEBUG`` is ``False`` (or, better, use another variable
            indicating a production vs. development environment).

        node_modules_path (unicode):
            The path to the loal :file:`node_modules` directory for the
            project.

        static_root (unicode):
            The value of the ``settings.STATIC_ROOT``. This must be provided
            explicitly, since :file:`settings.py` is likely the module
            calling this.

        javascript_bundles (list of dict, optional):
            A list of JavaScript bundle packages for Pipeline to handle.

        stylesheet_bundles (list of dict, optional):
            A list of stylesheet bundle packages for Pipeline to handle.

        compilers (list of unicode, optional):
            A list of compilers to use for static media.

        babel_extra_plugins (list of unicode, optional):
            A list of additional Babel plugins to enable.

        babel_extra_args (list of unicode, optional):
            Extra command line arguments to pass to Babel.

        less_extra_args (list of unicode, optional):
            Extra command line arguments to pass to LessCSS.

        validate_paths (bool, optional):
            Whether to validate any expected paths to binary files.

            It's recommended to set this based on ``DEBUG``, or another
            variable indicating a production vs. development environment.

            If the :envvar:`DJBLETS_SKIP_PIPELINE_VALIDATION` environment
            variable is set to ``1``, then this will be forced to ``False``.
            This is primarily used for packaging building.
    """
    bin_path = os.path.join(node_modules_path, '.bin')
    babel_bin_path = os.path.join(bin_path, 'babel')
    lessc_bin_path = os.path.join(bin_path, 'lessc')
    uglifyjs_bin_path = os.path.join(bin_path, 'uglifyjs')

    if (validate_paths and
        os.environ.get('DJBLETS_SKIP_PIPELINE_VALIDATION') != '1'):
        if not os.path.exists(node_modules_path):
            raise ImproperlyConfigured(
                'node_modules could not be found at %s'
                % node_modules_path)

        if not os.path.exists(babel_bin_path):
            raise ImproperlyConfigured(
                'The babel binary could not be found at %s'
                % babel_bin_path)

        if not os.path.exists(lessc_bin_path):
            raise ImproperlyConfigured(
                'The lessc binary could not be found at %s'
                % lessc_bin_path)

        if not os.path.exists(uglifyjs_bin_path):
            raise ImproperlyConfigured(
                'The uglifyjs binary could not be found at %s'
                % uglifyjs_bin_path)

    os.environ[str('NODE_PATH')] = force_str(node_modules_path)

    babel_plugins = [
        'dedent',
        'django-gettext',
    ] + babel_extra_plugins

    return {
        'PIPELINE_ENABLED': bool(pipeline_enabled),
        'COMPILERS': compilers,
        'CSS_COMPRESSOR': None,
        'JS_COMPRESSOR': 'pipeline.compressors.uglifyjs.UglifyJSCompressor',
        'JAVASCRIPT': javascript_bundles,
        'STYLESHEETS': stylesheet_bundles,
        'BABEL_BINARY': babel_bin_path,
        'BABEL_ARGUMENTS': [
            '--presets', '@babel/preset-env',
            '--plugins', ','.join(babel_plugins),
            '-s', 'true',
        ] + babel_extra_args,
        'LESS_BINARY': lessc_bin_path,
        'LESS_ARGUMENTS': [
            '--include-path=%s' % static_root,
            '--no-color',
            '--source-map',
            '--js',
            '--plugin=@beanbag/less-plugin-autoprefix',
        ] + less_extra_args,
        'UGLIFYJS_BINARY': uglifyjs_bin_path,
    }



"""Tests for WebAPI signals."""

from datetime import timedelta

from django.contrib.auth.models import User
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import HttpResponse
from django.test.client import RequestFactory
from django.test.utils import override_settings
from django.utils import timezone
from kgb import SpyAgency

from djblets.secrets.token_generators.vendor_checksum import \
    VendorChecksumTokenGenerator
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin
from djblets.webapi.auth.backends.api_tokens import (TokenAuthBackendMixin,
                                                     WebAPITokenAuthBackend)
from djblets.webapi.signals import (webapi_token_created,
                                    webapi_token_expired,
                                    webapi_token_updated)
from djblets.webapi.tests.test_api_token import WebAPIToken


class MyTestTokenAuthBackend(TokenAuthBackendMixin):
    """Mock Token Auth Backend for testing purposes."""

    api_token_model = WebAPIToken


@override_settings(AUTHENTICATION_BACKENDS=(
    'djblets.webapi.tests.test_signals.MyTestTokenAuthBackend',
))
class WebAPITokenSignalsTests(SpyAgency, TestModelsLoaderMixin, TestCase):
    """Tests for WebAPI token signals."""

    tests_app = 'djblets.webapi.tests'

    def setUp(self):
        super(WebAPITokenSignalsTests, self).setUp()

        self.user = User.objects.create_user(username='testuser',
                                             email='test@example.com')
        self.token_generator_id = \
            VendorChecksumTokenGenerator.token_generator_id
        self.token_info = {'token_type': 'test'}

        self.api_token_auth_backend = WebAPITokenAuthBackend()
        self.middleware = SessionMiddleware(lambda request: HttpResponse(''))

    def test_webapi_token_created(self):
        """Testing WebAPIToken.objects.generate_token() emits
        webapi_token_created
        """
        def on_webapi_token_created(auto_generated, **kwargs):
            self.assertFalse(auto_generated)

        webapi_token_created.connect(on_webapi_token_created)

        try:
            self.spy_on(on_webapi_token_created)
            WebAPIToken.objects.generate_token(
                self.user,
                token_generator_id=self.token_generator_id,
                token_info=self.token_info)
            self.assertTrue(on_webapi_token_created.spy.called)
        finally:
            webapi_token_created.disconnect(on_webapi_token_created)

    def test_webapi_token_created_autogenerated(self):
        """Testing WebAPIToken.objects.generate_token() with
        auto_generated=True emits webapi_token_created with
        auto_generated=True
        """
        def on_webapi_token_created(auto_generated, **kwargs):
            self.assertTrue(auto_generated)

        webapi_token_created.connect(on_webapi_token_created)

        try:
            self.spy_on(on_webapi_token_created)
            WebAPIToken.objects.generate_token(
                self.user,
                auto_generated=True,
                token_generator_id=self.token_generator_id,
                token_info=self.token_info)
            self.assertTrue(on_webapi_token_created.spy.called)
        finally:
            webapi_token_created.disconnect(on_webapi_token_created)

    def test_webapi_token_expired(self):
        """Testing WebAPITokenAuthBackend.authenticate() emits
        webapi_token_expired when authenticating with an expired token
        """
        def on_webapi_token_expired(**kwargs):
            pass

        request = RequestFactory().get('/')
        request.user = User()
        self.middleware(request)

        webapi_token_expired.connect(on_webapi_token_expired)

        try:
            self.spy_on(on_webapi_token_expired)

            # Testing that no signal is emitted when authenticating with
            # a non expired token.
            token = WebAPIToken.objects.generate_token(
                self.user,
                token_generator_id=self.token_generator_id,
                token_info=self.token_info)
            request.META['HTTP_AUTHORIZATION'] = 'token %s' % token.token

            self.api_token_auth_backend.authenticate(request)

            self.assertFalse(on_webapi_token_expired.spy.called)

            on_webapi_token_expired.spy.reset_calls()

            # Testing that a signal is emitted when authenticating with
            # an expired token.
            expired_token = WebAPIToken.objects.generate_token(
                self.user,
                expires=(timezone.now() - timedelta(days=1)),
                token_generator_id=self.token_generator_id,
                token_info=self.token_info)
            request.META['HTTP_AUTHORIZATION'] = ('token %s'
                                                  % expired_token.token)

            self.api_token_auth_backend.authenticate(request)

            self.assertTrue(on_webapi_token_expired.spy.called)
        finally:
            webapi_token_expired.disconnect(on_webapi_token_expired)

    def test_webapi_token_updated(self):
        """Testing WebAPIToken.save() emits webapi_token_updated"""
        def on_webapi_token_created(**kwargs):
            pass

        def on_webapi_token_updated(**kwargs):
            pass

        webapi_token_created.connect(on_webapi_token_created)
        webapi_token_updated.connect(on_webapi_token_updated)

        try:
            self.spy_on(on_webapi_token_created)
            self.spy_on(on_webapi_token_updated)

            token = WebAPIToken.objects.generate_token(
                self.user,
                token_generator_id=self.token_generator_id,
                token_info=self.token_info)

            self.assertTrue(on_webapi_token_created.spy.called)
            self.assertFalse(on_webapi_token_updated.spy.called)

            on_webapi_token_created.spy.reset_calls()

            token.save()

            self.assertFalse(on_webapi_token_created.spy.called)
            self.assertTrue(on_webapi_token_updated.spy.called)
        finally:
            webapi_token_created.disconnect(on_webapi_token_created)
            webapi_token_updated.disconnect(on_webapi_token_updated)



"""Unit tests for djblets.util.dates."""

from datetime import datetime

import pytz

from djblets.testing.testcases import TestCase
from djblets.util.dates import (get_latest_timestamp, get_tz_aware_utcnow,
                                http_date)


class DatesTests(TestCase):
    """Unit tests for djblets.util.dates."""

    def test_http_date_with_datetime(self):
        """Testing http_date with datetime"""
        date_time = datetime(2016, 8, 26, 3, 3, 26, 123456)
        self.assertEqual(http_date(date_time),
                         'Fri, 26 Aug 2016 03:03:26 GMT')

    def test_http_date_with_date_string(self):
        """Testing http_date with date string"""
        date = '20/06/2016'
        self.assertEqual(http_date(date), date)

    def test_http_date_with_unix_timestamp(self):
        """Testing http_date with unix timestamp"""
        unix_timestamp = '1466424000'
        self.assertEqual(http_date(unix_timestamp), unix_timestamp)

    def test_get_latest_timestamp_with_empty_list(self):
        """Testing get_latest_timestamp without any timestamps in the list"""
        self.assertIsNone(get_latest_timestamp([]))

    def test_get_latest_timestamp_with_jumbled_list(self):
        """Testing get_latest_timestamp with unsorted date time
        list.
         """
        self.assertEqual(
            get_latest_timestamp([
                '1453204800',
                '1466337600',
                '1466424000'
            ]),
            '1466424000')

    def test_date_time_is_in_utc(self):
        """Testing get_tz_aware_utcnow returns UTC time."""
        utc_time = get_tz_aware_utcnow()
        self.assertEqual(utc_time.tzinfo, pytz.utc)



"""Template tags for working with features."""

from django import template
from django.template.base import (Node, NodeList, TemplateSyntaxError,
                                  token_kwargs)

from djblets.features.registry import get_features_registry


register = template.Library()


class IfFeatureNode(Node):
    """Template node for feature-based if statements.

    This works mostly like a standard ``{% if %}`` tag, checking whether the
    given feature is enabled and rendering the content between it and the
    else/end tags only if matching the desired state.

    This supports a ``{% else %}``, to allow rendering content if the feature
    does not match the desired state.

    This is used by both the
    :py:class:`{% if_feature_enabled %} <if_feature_enabled>` and
    :py:class:`{% if_feature_disabled %} <if_feature_disabled>` tags.
    """

    child_nodelists = ('nodelist_true', 'nodelist_false')

    def __init__(self, nodelist_enabled, nodelist_disabled, feature_id,
                 extra_kwargs):
        """Initialize the template node.

        Args:
            nodelist_enabled (django.template.NodeList):
                The nodelist to render if the feature is enabled.

            nodelist_disabled (django.template.NodeList):
                The nodelist to render if the feature is disabled.

            feature_id (unicode):
                The ID of the feature to check.

            extra_kwargs (dict):
                Extra keyword arguments to pass to
                :py:meth:`Feature.is_enabled()
                <djblets.features.feature.Feature.is_enabled>`.
        """

        self.nodelist_enabled = nodelist_enabled
        self.nodelist_disabled = nodelist_disabled
        self.feature_id = feature_id
        self.extra_kwargs = extra_kwargs

    def __repr__(self):
        """Return a representation of the node.

        This is mostly used for debugging output.

        Returns:
            unicode:
            A representation of this node.
        """
        return '<IfFeatureNode for %s>' % self.feature_id

    def render(self, context):
        """Render the node.

        This will determine if the feature is enabled or disabled, and render
        the appropriate list of nodes to a string.

        Args:
            context (django.template.Context):
                The context provided by the template.

        Returns:
            unicode:
            The rendered content as a string.
        """
        feature_id = self.feature_id.resolve(context, True)
        extra_kwargs = {
            key: value.resolve(context)
            for key, value in self.extra_kwargs.items()
        }

        feature = get_features_registry().get_feature(feature_id)

        if feature:
            enabled = feature.is_enabled(request=context.get('request'),
                                         **extra_kwargs)
        else:
            enabled = False

        if enabled:
            return self.nodelist_enabled.render(context)
        else:
            return self.nodelist_disabled.render(context)


def _if_feature(parser, token, enabled_first):
    """Common implementation for feature-based if statements.

    This constructs a :py:class:`IfFeatureNode` for the consuming template tag,
    allowing for "if" and "if not" checks.

    Args:
        parser (django.template.Parser):
            The parser being used to parse this template tag.

        token (django.template.Token):
            The token representing this template tag.

        enabled_first (bool):
            If ``True``, this behaves as an "if enabled" check.
            If ``False``, this behaves as a "if disabled' check.

    Returns:
        IfFeatureNode:
        The feature checker node to use for the template.
    """
    bits = token.split_contents()
    tag = bits[0]
    end_tag = 'end%s' % tag

    if len(bits) < 2:
        raise TemplateSyntaxError('%r requires a feature ID argument'
                                  % tag)

    nodelist_1 = parser.parse(('else', end_tag))
    token = parser.next_token()

    if token.contents == 'else':
        nodelist_2 = parser.parse((end_tag,))
        parser.delete_first_token()
    else:
        nodelist_2 = NodeList()

    if enabled_first:
        nodelist_enabled = nodelist_1
        nodelist_disabled = nodelist_2
    else:
        nodelist_disabled = nodelist_1
        nodelist_enabled = nodelist_2

    feature_id = parser.compile_filter(bits[1])
    remaining_bits = bits[2:]
    extra_kwargs = token_kwargs(remaining_bits, parser)

    if remaining_bits:
        raise TemplateSyntaxError('%r received an invalid token: %r'
                                  % (tag, remaining_bits[0]))

    return IfFeatureNode(nodelist_enabled, nodelist_disabled, feature_id,
                         extra_kwargs)


@register.tag
def if_feature_enabled(parser, token):
    """Render content only if a feature is enabled.

    This works mostly like a standard ``{% if %}`` tag, checking if the
    given feature is enabled before rendering the content between it and the
    else or end tags.

    This supports a ``{% else %}``, to allow rendering alternative content if
    the feature is disabled instead.

    It also accepts additional keyword arguments that can be passed to
    :py:meth:`Feature.is_enabled()
    <djblets.features.feature.Feature.is_enabled>`.

    Args:
        parser (django.template.Parser):
            The parser being used to parse this template tag.

        token (django.template.Token):
            The token representing this template tag.

    Returns:
        IfFeatureNode:
        The feature checker node to use for the template.

    Example:
        .. code-block:: html+django

           {% if_feature_enabled "my-feature" user=request.user %}
           This will only render if the feature is enabled for the user.
           {% else %}
           This will only render if the feature is disabled for the user.
           {% endif_feature_enabled %}
    """
    return _if_feature(parser, token, enabled_first=True)


@register.tag
def if_feature_disabled(parser, token):
    """Render content only if a feature is disabled.

    This works mostly like a standard ``{% if %}`` tag, checking if the
    given feature is disabled before rendering the content between it and the
    else or end tags.

    This supports a ``{% else %}``, to allow rendering alternative content if
    the feature is enabled instead.

    It also accepts additional keyword arguments that can be passed to
    :py:meth:`Feature.is_enabled()
    <djblets.features.feature.Feature.is_enabled>`.

    Args:
        parser (django.template.Parser):
            The parser being used to parse this template tag.

        token (django.template.Token):
            The token representing this template tag.

    Returns:
        IfFeatureNode:
        The feature checker node to use for the template.

    Example:
        .. code-block:: html+django

           {% if_feature_disabled "my-feature" user=request.user %}
           This will only render if the feature is disabled for the user.
           {% else %}
           This will only render if the feature is enabled for the user.
           {% endif_feature_disabled %}
    """
    return _if_feature(parser, token, enabled_first=False)



"""Authentication-related views.

The bundled views help with common authentication-related tasks not otherwise
provided by Django. At the moment, there is only support here for registration.
"""

from django.contrib import auth
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.views.decorators.csrf import csrf_protect

from djblets.auth.forms import RegistrationForm
from djblets.auth.signals import user_registered
from djblets.auth.util import validate_test_cookie


@csrf_protect
def register(request, next_page, form_class=RegistrationForm,
             extra_context=None, initial_values=None,
             form_kwargs=None, template_name="accounts/register.html"):
    """Handle registration of a new user.

    This works along with :py:class:`djblets.auth.forms.RegistrationForm`
    to register a new user. It will display a registration form, validate
    the user's new information, and then log them in.

    The registration form, next page, and context can all be customized by
    the caller.

    Args:
        request (HttpRequest):
            The HTTP request from the client.

        next_page (unicode):
            The URL to navigate to once registration is successful.

        form_class (Form subclass):
            The form that will handle registration, field validation, and
            creation of the user.

        extra_context (dict):
            Extra context variables to pass to the template when rendering.

        initial_values (dict):
            Initial values to set on the form when it is rendered.

        form_kwargs (dict):
            Additional keyword arguments to pass to the form class during
            instantiation.

        template_name (unicode):
            The name of the template containing the registration form.

    Returns:
        HttpResponse: The page's rendered response or redirect.
    """
    if initial_values is None:
        initial_values = {}

    if form_kwargs is None:
        form_kwargs = {}

    if request.method == 'POST':
        form = form_class(data=request.POST, request=request, **form_kwargs)
        form.full_clean()
        validate_test_cookie(form, request)

        if form.is_valid():
            user = form.save()
            if user:
                user = auth.authenticate(
                    username=form.cleaned_data['username'],
                    password=form.cleaned_data['password1'])
                assert user
                auth.login(request, user)
                try:
                    request.session.delete_test_cookie()
                except KeyError:
                    # Do nothing
                    pass

                # Other components can listen to this signal to
                # perform additional tasks when a new user registers
                user_registered.send(sender=None, user=request.user)

                return HttpResponseRedirect(next_page)
    else:
        form = form_class(initial=initial_values, request=request,
                          **form_kwargs)

    request.session.set_test_cookie()

    context = {
        'form': form,
    }

    if extra_context:
        context.update(extra_context)

    return render(request, template_name, context)



"""Unit tests for djblets.forms.fields.ListEditDictionaryField."""

from djblets.forms.fields import ListEditDictionaryField
from djblets.testing.testcases import TestCase


class ListEditDictionaryFieldTests(TestCase):
    """Unit tests for djblets.forms.fields.ListEditDictionaryField."""

    def test_prepare_value(self):
        """Testing ListEditDictionaryField.prepare_value"""
        field = ListEditDictionaryField()

        self.assertEqual(
            field.prepare_value({1: 'foo',
                                 2: 'bar',
                                 3: 'baz'}),
            [(1, 'foo'), (2, 'bar'), (3, 'baz')])

    def test_prepare_value_with_null_value(self):
        """Testing ListEditDictionaryField.prepare_value with a null value"""
        field = ListEditDictionaryField()

        self.assertEqual(
            field.prepare_value(None),
            [])

    def test_prepare_value_with_list_value(self):
        """Testing ListEditDictionaryField.prepare_value with a list"""
        field = ListEditDictionaryField()

        self.assertEqual(
            field.prepare_value([(1, 'foo'), (2, 'bar'), (3, 'baz')]),
            [(1, 'foo'), (2, 'bar'), (3, 'baz')])

    def test_to_python(self):
        """Testing ListEditDictionaryField.to_python"""
        field = ListEditDictionaryField()

        self.assertEqual(
            field.to_python([(1, 'foo'), (2, 'bar'), (3, 'baz')]),
            {1: 'foo',
             2: 'bar',
             3: 'baz'})



"""Unit tests for djblets.util.http."""

from django.http import HttpRequest, HttpResponse, HttpResponseNotModified

from djblets.testing.testcases import TestCase
from djblets.util.http import (build_not_modified_from_response,
                               get_http_accept_lists,
                               get_http_requested_mimetype,
                               is_mimetype_a)


class HttpTests(TestCase):
    """Unit tests for djblets.util.http."""

    def setUp(self):
        self.request = HttpRequest()
        self.request.META['HTTP_ACCEPT'] = \
            'application/json;q=0.5,application/xml,text/plain;q=0.0,*/*;q=0.0'

    def test_http_accept_lists(self):
        """Testing get_http_accept_lists"""
        acceptable_mimetypes, unacceptable_mimetypes = \
            get_http_accept_lists(self.request)

        self.assertEqual(acceptable_mimetypes,
                         ['application/xml', 'application/json'])
        self.assertEqual(unacceptable_mimetypes, ['text/plain', '*/*'])

    def test_get_requested_mimetype_with_supported_mimetype(self):
        """Testing get_requested_mimetype with supported mimetype"""
        self.assertEqual(
            get_http_requested_mimetype(self.request, ['foo/bar',
                                                       'application/json']),
            'application/json')
        self.assertEqual(
            get_http_requested_mimetype(self.request, ['application/xml']),
            'application/xml')
        self.assertEqual(
            get_http_requested_mimetype(self.request, ['application/json',
                                                       'application/xml']),
            'application/xml')

    def test_get_requested_mimetype_with_no_consensus(self):
        """Testing get_requested_mimetype with no consensus between client and
        server
        """
        self.request.META['HTTP_ACCEPT'] = ('text/html,application/xhtml+xml,'
                                            'application/xml;q=0.9,*/*;q=0.8')

        self.assertEqual(
            get_http_requested_mimetype(self.request, ['application/json',
                                                       'application/x-foo']),
            'application/json')

    def test_get_requested_mimetype_with_wildcard_supported_mimetype(self):
        """Testing get_requested_mimetype with supported */* mimetype"""
        self.request.META['HTTP_ACCEPT'] = '*/*'
        self.assertEqual(
            get_http_requested_mimetype(self.request, ['application/json',
                                                       'application/xml']),
            'application/json')

    def test_get_requested_mimetype_with_unsupported_mimetype(self):
        """Testing get_requested_mimetype with unsupported mimetype"""
        self.assertIsNone(get_http_requested_mimetype(self.request,
                                                      ['text/plain']))
        self.assertIsNone(get_http_requested_mimetype(self.request,
                                                      ['foo/bar']))

    def test_is_mimetype_a(self):
        """Testing is_mimetype_a"""
        self.assertTrue(is_mimetype_a('application/json',
                                      'application/json'))
        self.assertTrue(is_mimetype_a('application/vnd.foo+json',
                                      'application/json'))
        self.assertFalse(is_mimetype_a('application/xml',
                                       'application/json'))
        self.assertFalse(is_mimetype_a('foo/vnd.bar+json',
                                       'application/json'))

    def test_build_not_modified_from_response_with_headers(self):
        """Testing build_not_modified_from_response with copyable headers
        in response
        """
        response = HttpResponse()
        response['Cache-Control'] = 'max-age=10000'
        response['Content-Location'] = '/foo/'
        response['Date'] = 'Mon, 12 Mar 2018 20:17:00 GMT'
        response['ETag'] = 'abc123'
        response['Expires'] = 'Wed, 14 Mar 2018 13:00:00 GMT'
        response['Vary'] = 'Cookie'

        new_response = build_not_modified_from_response(response)
        self.assertIsInstance(new_response, HttpResponseNotModified)
        self.assertEqual(new_response['Cache-Control'], 'max-age=10000')
        self.assertEqual(new_response['Content-Location'], '/foo/')
        self.assertEqual(new_response['Date'], 'Mon, 12 Mar 2018 20:17:00 GMT')
        self.assertEqual(new_response['ETag'], 'abc123')
        self.assertEqual(new_response['Expires'],
                         'Wed, 14 Mar 2018 13:00:00 GMT')
        self.assertEqual(new_response['Vary'], 'Cookie')

    def test_build_not_modified_from_response_without_headers(self):
        """Testing build_not_modified_from_response without copyable headers
        in response
        """
        new_response = build_not_modified_from_response(HttpResponse())
        self.assertIsInstance(new_response, HttpResponseNotModified)
        self.assertNotIn('Cache-Control', new_response)
        self.assertNotIn('Content-Location', new_response)
        self.assertNotIn('Date', new_response)
        self.assertNotIn('ETag', new_response)
        self.assertNotIn('Expires', new_response)
        self.assertNotIn('Vary', new_response)



"""Error classes for extensions."""

from django.utils.translation import gettext as _


class EnablingExtensionError(Exception):
    """An extension could not be enabled."""

    def __init__(self, message, load_error=None, needs_reload=False):
        """Initialize the error.

        Args:
            message (unicode):
                The detailed error message.

            load_error (unicode, optional):
                An exception from the attempt to enable the extension, or
                other useful information to display to the user to help
                diagnose the problem.

            needs_reload (bool, optional):
                Whether fixing this error requires reloading the extension.
        """
        super(EnablingExtensionError, self).__init__(message)

        self.load_error = load_error
        self.needs_reload = needs_reload


class DisablingExtensionError(Exception):
    """An extension could not be disabled."""
    pass


class InstallExtensionError(Exception):
    """An extension could not be installed."""
    def __init__(self, message, load_error=None):
        self.message = message
        self.load_error = load_error


class InstallExtensionMediaError(InstallExtensionError):
    """An error indicating that extension media files could not be installed.
    """


class InvalidExtensionError(Exception):
    """An extension does not exist."""
    def __init__(self, extension_id):
        super(InvalidExtensionError, self).__init__()
        self.message = _("Cannot find extension with id %s") % extension_id



"""Extension hooks for registering integrations."""

import logging

from djblets.extensions.hooks import ExtensionHook
from djblets.integrations.errors import IntegrationError
from djblets.integrations.mixins import NeedsIntegrationManagerMixin


class BaseIntegrationHook(NeedsIntegrationManagerMixin, ExtensionHook):
    """Base class for an extension hook for integration management.

    Integrations enable any supporting application to connect with third-party
    services in specialized ways. This class makes it easy to register new
    integrations on an extension, binding their lifecycles to that of the
    extension.

    This is not intended to be used directly by a caller. Instead, a project
    offering integrations should subclass this and provide an implementation
    of :py:meth:`get_integration_manager`.
    """

    def initialize(self, integration_cls):
        """Initialize the hook.

        The provided integration will be registered with the integration
        manager.

        Args:
            integration_cls (type):
                The integration class to register.
        """
        self.integration_cls = integration_cls

        manager = self.get_integration_manager()

        try:
            manager.register_integration_class(self.integration_cls)
        except IntegrationError as e:
            logging.exception('%s could not register integration class '
                              '%s with %s: %s',
                              self.extension.__class__.__name__,
                              self.__class__.__name__,
                              self.integration_cls.__name__,
                              e)

    def shutdown(self):
        """Shut down the hook.

        The integration registered in this hook will be unregistered.
        """
        manager = self.get_integration_manager()

        if manager.enabled:
            try:
                manager.unregister_integration_class(self.integration_cls)
            except IntegrationError as e:
                logging.exception('%s could not unregister integration class '
                                  '%s with %s: %s',
                                  self.extension.__class__.__name__,
                                  self.__class__.__name__,
                                  self.integration_cls.__name__,
                                  e)



"""Unit tests for djblets.configforms.registry.ConfigPageRegistry."""

from kgb import SpyAgency

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.mixins import DynamicConfigPageMixin
from djblets.configforms.pages import ConfigPage
from djblets.configforms.registry import ConfigPageRegistry
from djblets.registries.errors import RegistrationError
from djblets.testing.testcases import TestCase


class DynamicConfigPage(DynamicConfigPageMixin, ConfigPage):
    pass


class MyTestFormA(ConfigPageForm):
    form_id = 'form-a'


class MyTestFormB(ConfigPageForm):
    form_id = 'form-b'


class MyTestPageOne(DynamicConfigPage):
    page_id = 'page-b'
    form_classes = [MyTestFormB]


class MyTestPageTwo(DynamicConfigPage):
    page_id = 'page-ab'
    form_classes = [MyTestFormA, MyTestFormB]


class ConfigPageRegistryTests(SpyAgency, TestCase):
    """Unit tests for djblets.configforms.registry.ConfigPageRegistry."""

    @classmethod
    def setUpClass(cls):
        super(ConfigPageRegistryTests, cls).setUpClass()
        cls.registry = ConfigPageRegistry()
        DynamicConfigPage.registry = cls.registry

    def tearDown(self):
        super(ConfigPageRegistryTests, self).tearDown()

        self.registry.reset()
        MyTestPageOne.form_classes = [MyTestFormB]
        MyTestPageTwo.formClasses = [MyTestFormA, MyTestFormB]

    def test_reset(self):
        """Testing ConfigPageRegistry.reset"""
        self.assertSetEqual(set(self.registry), set())
        self.assertSetEqual(set(self.registry._forms), set())
        self.registry.register(MyTestPageOne)
        self.registry.reset()

        self.assertSetEqual(set(self.registry), set())
        self.assertSetEqual(set(self.registry._forms), set())

    def test_partial_register_page(self):
        """Testing ConfigPageRegistry.register when the registration is
        partially successful
        """
        self.registry.register(MyTestPageOne)

        with self.assertRaises(RegistrationError):
            self.registry.register(MyTestPageTwo)

        self.assertEqual(set(self.registry), {MyTestPageOne})

    def test_register_page_with_duplicate(self):
        """Testing ConfigPageRegistry.register when registering a page that
        contains pre-registered forms
        """
        class MyTestPage(DynamicConfigPage):
            page_id = 'test-page'
            form_classes = [MyTestFormB]

        self.registry.register(MyTestPageOne)

        with self.assertRaises(RegistrationError):
            self.registry.register(MyTestPage)

    def test_add_form_to_page(self):
        """Testing ConfigPageRegistry.add_form_to_page"""
        self.registry.register(MyTestPageOne)
        self.registry.add_form_to_page(MyTestPageOne, MyTestFormA)

        self.assertSetEqual(set(MyTestPageOne.form_classes),
                            {MyTestFormA, MyTestFormB})
        self.assertSetEqual(set(self.registry._forms),
                            {MyTestFormA, MyTestFormB})

    def test_remove_form_from_page(self):
        """Testing ConfigPageRegistry.remove_form_from_page"""
        self.registry.register(MyTestPageOne)
        self.registry.remove_form_from_page(MyTestPageOne, MyTestFormB)

        self.assertListEqual(list(MyTestPageOne.form_classes), [])
        self.assertListEqual(list(self.registry._forms), [])

    def test_add_duplicate_form_to_page(self):
        """Testing ConfigPageRegistry.add_form_to_page with a duplicate form"""
        self.registry.register(MyTestPageOne)

        with self.assertRaises(RegistrationError):
            self.registry.add_form_to_page(MyTestPageOne, MyTestFormB)

        self.assertEqual(MyTestPageOne.form_classes, [MyTestFormB])

    def test_default_form_classes(self):
        """Testing DynamicConfigPageMixin._default_form_classes persistence"""
        self.registry.register(MyTestPageOne)
        self.assertListEqual(MyTestPageOne.form_classes, [MyTestFormB])
        self.registry.add_form_to_page(MyTestPageOne, MyTestFormA)
        self.assertSetEqual(set(MyTestPageOne.form_classes),
                            {MyTestFormA, MyTestFormB})

        self.registry.unregister(MyTestPageOne)
        self.assertListEqual(MyTestPageOne.form_classes, [])

        self.registry.register(MyTestPageOne)
        self.assertListEqual(MyTestPageOne.form_classes, [MyTestFormB])

    def test_empty_default_form_classes_for_page(self):
        """Testing DynamicConfigPageMixin._default_form_classes with no form
        classes
        """
        class MyTestPage(DynamicConfigPage):
            page_id = 'test-page'
            page_title = 'Test Page'

        self.registry.register(MyTestPage)
        self.assertListEqual(MyTestPage.form_classes, [])

        self.registry.add_form_to_page(MyTestPage, MyTestFormA)
        self.assertListEqual(MyTestPage.form_classes, [MyTestFormA])

        self.registry.unregister(MyTestPage)
        self.assertListEqual(MyTestPage.form_classes, [])

        self.registry.register(MyTestPage)
        self.assertListEqual(MyTestPage.form_classes, [])

    def test_add_form_to_page_populate(self):
        """Testing ConfigPageRegistry.add_form_to_page populates itself and the
         ConfigPageFormRegistry
         """
        class TestRegistry(ConfigPageRegistry):
            def get_defaults(self):
                yield MyTestPageOne

        registry = TestRegistry()

        self.spy_on(registry.populate)
        self.spy_on(registry._forms.populate)

        registry.add_form_to_page(MyTestPageOne, MyTestFormA)

        self.assertTrue(registry.populate.spy.called)
        self.assertTrue(registry._forms.populate.spy.called)
        self.assertEqual(set(registry._forms), {MyTestFormA, MyTestFormB})

    def test_add_form_to_page_populate_duplicate(self):
        """Testing ConfigPageRegistry.add_form_to_page with an already
        registered page raises an error due to population
        """
        class TestRegistry(ConfigPageRegistry):
            def get_defaults(self):
                yield MyTestPageOne

        registry = TestRegistry()

        with self.assertRaises(registry.already_registered_error_class):
            registry.add_form_to_page(MyTestPageOne, MyTestFormB)

        self.assertEqual(set(registry._forms), {MyTestFormB})

    def test_remove_form_from_page_populate(self):
        """Testing ConfigPageRegistry.remove_form_from_page populates itself
        and the ConfigPgaeFormRegistry
        """
        class TestRegistry(ConfigPageRegistry):
            def get_defaults(self):
                yield MyTestPageOne

        registry = TestRegistry()

        self.spy_on(registry.populate)
        self.spy_on(registry._forms.populate)

        registry.remove_form_from_page(MyTestPageOne, MyTestFormB)

        self.assertTrue(registry.populate.spy.called)
        self.assertTrue(registry._forms.populate.spy.called)
        self.assertEqual(set(registry._forms), set())

    def test_remove_form_from_page_populate_unregistered(self):
        """Testing ConfigPageForm.remove_form_from_page with an unregistered
        page raises an error due to population
        """
        class TestRegistry(ConfigPageRegistry):
            def get_defaults(self):
                yield MyTestPageOne

        registry = TestRegistry()

        with self.assertRaises(registry.lookup_error_class):
            registry.remove_form_from_page(MyTestPageOne, MyTestFormA)

        self.assertEqual(set(registry._forms), {MyTestFormB})



"""Unit tests for djblets.configforms.pages.ConfigPage."""

import re

from django.contrib.auth.models import User
from django.test.client import RequestFactory

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.pages import ConfigPage
from djblets.configforms.views import ConfigPagesView
from djblets.testing.testcases import TestCase


class MyTestForm1(ConfigPageForm):
    form_id = 'my-form-1'
    form_title = 'Form 1'


class MyTestForm2(ConfigPageForm):
    form_id = 'my-form-2'
    form_title = 'Form 2'


class MyTestForm3(ConfigPageForm):
    form_id = 'my-form-3'
    form_title = 'Form 3'

    def is_visible(self):
        return False


class MyTestPage(ConfigPage):
    page_id = 'my-page'
    form_classes = [MyTestForm1, MyTestForm2, MyTestForm3]


class ConfigPageTests(TestCase):
    """Unit tests for djblets.configforms.pages.ConfigPage."""

    def setUp(self):
        super(ConfigPageTests, self).setUp()

        self.request = RequestFactory().request()
        self.user = User.objects.create_user(username='test-user',
                                             password='test-user')
        self.page = MyTestPage(ConfigPagesView, self.request, self.user)

    def test_initial_state(self):
        """Testing ConfigPage initial state"""
        self.assertEqual(len(self.page.forms), 2)
        self.assertIsInstance(self.page.forms[0], MyTestForm1)
        self.assertIsInstance(self.page.forms[1], MyTestForm2)

    def test_is_visible_with_visible_forms(self):
        """Testing ConfigPage.is_visible with visible forms"""
        self.assertTrue(self.page.is_visible())

    def test_is_visible_with_no_visible_forms(self):
        """Testing ConfigPage.is_visible without visible forms"""
        class MyTestPage(ConfigPage):
            page_id = 'my-page'
            form_classes = [MyTestForm3]

        page = MyTestPage(ConfigPagesView, self.request, self.user)
        self.assertFalse(page.is_visible())

    def test_render(self):
        """Testing ConfigPage.render"""
        # Filter out the CSRF token, since it's hard to match.
        rendered = re.sub(r'<input.+name="csrfmiddlewaretoken".*>',
                          '',
                          self.page.render())

        self.assertHTMLEqual(
            rendered,
            '<div class="box-container">'
            ' <div class="box">'
            '  <div class="box-inner">'
            '   <div class="box-head">'
            '    <h1 class="box-title">Form 1</h1>'
            '   </div>'
            '   <div class="box-main box-foot">'
            '    <form method="post" action=".#my-page" id="form_my-form-1">'
            '     <div class="djblets-l-config-forms-container">'
            '      <input id="id_form_target" name="form_target"'
            '             type="hidden" value="my-form-1">'
            '      <input type="submit" class="btn" value="Save">'
            '     </div>'
            '    </form>'
            '   </div>'
            '  </div>'
            ' </div>'
            '</div>'
            '<div class="box-container">'
            ' <div class="box">'
            '  <div class="box-inner">'
            '   <div class="box-head">'
            '    <h1 class="box-title">Form 2</h1>'
            '   </div>'
            '   <div class="box-main box-foot">'
            '    <form method="post" action=".#my-page" id="form_my-form-2">'
            '     <div class="djblets-l-config-forms-container">'
            '      <input id="id_form_target" name="form_target"'
            '             type="hidden" value="my-form-2">'
            '      <input type="submit" class="btn" value="Save">'
            '     </div>'
            '    </form>'
            '   </div>'
            '  </div>'
            ' </div>'
            '</div>')



"""Unit tests for djblets.util.properties."""

from djblets.testing.testcases import TestCase
from djblets.util.properties import AliasProperty, TypedProperty


class MyDeprecationWarning(DeprecationWarning):
    pass


class AliasPropertyTests(TestCase):
    """Unit tests for djblets.util.properties.AliasProperty."""

    def test_set(self):
        """Testing AliasProperty.__set__"""
        class MyObject(object):
            prop = AliasProperty('other_prop')
            other_prop = 100

        obj = MyObject()

        with self.assertNoWarnings():
            obj.prop = 42

        self.assertEqual(obj.other_prop, 42)

    def test_set_with_deprecated(self):
        """Testing AliasProperty.__set__ with deprecated=True"""
        class MyObject(object):
            prop = AliasProperty('other_prop',
                                 deprecated=True,
                                 deprecation_warning=MyDeprecationWarning)
            other_prop = 100

        obj = MyObject()

        expected_message = (
            'MyObject.prop is deprecated. Please set MyObject.other_prop '
            'instead.'
        )

        with self.assertWarns(MyDeprecationWarning, expected_message):
            obj.prop = 42

        self.assertEqual(obj.other_prop, 42)

    def test_set_with_convert_to_func(self):
        """Testing AliasProperty.__set__ with convert_to_func"""
        class MyObject(object):
            prop = AliasProperty('other_prop',
                                 convert_to_func=str)
            other_prop = '100'

        obj = MyObject()

        with self.assertNoWarnings():
            obj.prop = 42

        self.assertEqual(obj.other_prop, '42')

    def test_set_with_convert_to_func_and_none(self):
        """Testing AliasProperty.__set__ with convert_to_func and value=None"""
        class MyObject(object):
            prop = AliasProperty('other_prop',
                                 convert_to_func=str)
            other_prop = 100

        obj = MyObject()

        with self.assertNoWarnings():
            obj.prop = None

        self.assertIsNone(obj.other_prop)

    def test_get(self):
        """Testing AliasProperty.__get__"""
        class MyObject(object):
            prop = AliasProperty('other_prop')
            other_prop = 100

        obj = MyObject()

        with self.assertNoWarnings():
            self.assertEqual(obj.prop, 100)

    def test_get_with_deprecated(self):
        """Testing AliasProperty.__get__ with deprecated=True"""
        class MyObject(object):
            prop = AliasProperty('other_prop',
                                 deprecated=True,
                                 deprecation_warning=MyDeprecationWarning)
            other_prop = 100

        obj = MyObject()

        expected_message = (
            'MyObject.prop is deprecated. Please access MyObject.other_prop '
            'instead.'
        )

        with self.assertWarns(MyDeprecationWarning, expected_message):
            self.assertEqual(obj.prop, 100)

    def test_get_with_convert_from_func(self):
        """Testing AliasProperty.__get__ with convert_from_func"""
        class MyObject(object):
            prop = AliasProperty('other_prop',
                                 convert_from_func=str)
            other_prop = 100

        obj = MyObject()

        with self.assertNoWarnings():
            self.assertEqual(obj.prop, '100')

    def test_get_with_convert_from_func_and_none(self):
        """Testing AliasProperty.__get__ with convert_from_func and value=None
        """
        class MyObject(object):
            prop = AliasProperty('other_prop',
                                 convert_to_func=str)
            other_prop = None

        obj = MyObject()

        with self.assertNoWarnings():
            self.assertIsNone(obj.prop)


class TypedPropertyTests(TestCase):
    """Unit tests for djblets.util.properties.TypedProperty."""

    def test_with_valid_type(self):
        """Testing TypedProperty with valid type"""
        class MyObject(object):
            prop = TypedProperty((int, bool))

        obj = MyObject()

        obj.prop = 100
        self.assertEqual(obj.prop, 100)

        obj.prop = True
        self.assertIs(obj.prop, True)

    def test_with_invalid_type(self):
        """Testing TypedProperty with invalid type"""
        class MyObject(object):
            prop = TypedProperty((str,))

        obj = MyObject()

        expected_message = 'int (100) is not a valid type for this property.'

        with self.assertRaisesMessage(TypeError, expected_message):
            obj.prop = 100

    def test_set_with_none_and_allow_none_true(self):
        """Testing TypedProperty.__set__ with None value and allow_none=True"""
        class MyObject(object):
            prop = TypedProperty((int, bool), allow_none=True)

        obj = MyObject()
        obj.prop = None
        self.assertIsNone(obj.prop)

    def test_set_with_none_and_allow_none_false(self):
        """Testing TypedProperty.__set__ with None value and allow_none=False
        """
        class MyObject(object):
            prop = TypedProperty((int, bool), allow_none=False)

        obj = MyObject()

        expected_message = 'None is not a valid value for this property.'

        with self.assertRaisesMessage(TypeError, expected_message):
            obj.prop = None

    def test_get_with_default(self):
        """Testing TypedProperty.__get__ with default value"""
        class MyObject(object):
            prop = TypedProperty((int, bool), default=100)

        obj = MyObject()
        self.assertEqual(obj.prop, 100)



from django.conf import settings


def site_root(request):
    """
    Exposes a SITE_ROOT variable in templates. This assumes that the
    project has been configured with a SITE_ROOT settings variable and
    proper support for basing the installation in a subdirectory.
    """
    return {'SITE_ROOT': settings.SITE_ROOT}



"""Unit tests for djblets.privacy.pii."""

from django.http import QueryDict
from django.test.client import RequestFactory
from django.urls import ResolverMatch

from djblets.privacy.pii import (build_pii_safe_page_url,
                                 build_pii_safe_page_url_for_request)
from djblets.testing.testcases import TestCase


class BuildPIISafePageURLTests(TestCase):
    """Unit tests for build_pii_safe_page_url."""

    def test_with_pii_keyword_in_path(self):
        """Testing build_pii_safe_page_url_for_request with PII
        keyword-matching in URL path
        """
        self.assertEqual(
            build_pii_safe_page_url(
                url='/users/test/',
                url_kwargs={
                    'user': 'test',
                }),
            '/users/<REDACTED>/')

        self.assertEqual(
            build_pii_safe_page_url(
                url='/users/test/',
                url_kwargs={
                    'username': 'test',
                }),
            '/users/<REDACTED>/')

        self.assertEqual(
            build_pii_safe_page_url(
                url='/users/test/',
                url_kwargs={
                    'email': 'test',
                }),
            '/users/<REDACTED>/')

    def test_with_pii_keyword_in_querystring(self):
        """Testing build_pii_safe_page_url with PII
        keyword-matching in URL query string
        """
        self.assertEqual(
            build_pii_safe_page_url(
                url='/',
                query_dict=QueryDict('user=test')),
            '/?user=<REDACTED>')

        self.assertEqual(
            build_pii_safe_page_url(
                url='/',
                query_dict=QueryDict('username=test')),
            '/?username=<REDACTED>')

        self.assertEqual(
            build_pii_safe_page_url(
                url='/',
                query_dict=QueryDict('email=test')),
            '/?email=<REDACTED>')

    def test_with_pii_email_in_path(self):
        """Testing build_pii_safe_page_url with email-like value
        in URL path
        """
        self.assertEqual(
            build_pii_safe_page_url(
                url='/test/a@b.com/',
                url_kwargs={
                    'foo': 'a@b.com',
                }),
            '/test/<REDACTED>/')

    def test_with_pii_email_in_querystring(self):
        """Testing build_pii_safe_page_url with email-like value
        in URL query string
        """
        self.assertEqual(
            build_pii_safe_page_url(
                url='/',
                query_dict=QueryDict('foo=a@b.com')),
            '/?foo=<REDACTED>')

    def test_with_non_string_in_path(self):
        """Testing build_pii_safe_page_url with non-string value in URL path"""
        self.assertEqual(
            build_pii_safe_page_url(
                url='/test/',
                url_kwargs={
                    'type': object,
                }),
            '/test/')

    def test_with_no_pii(self):
        """Testing build_pii_safe_page_url with no PII"""
        self.assertEqual(
            build_pii_safe_page_url(
                url='/groups/test/',
                url_kwargs={
                    'groupname': 'test',
                }),
            '/groups/test/')

    def test_with_no_keywords(self):
        """Testing build_pii_safe_page_url with no keywords in URL
        pattern
        """
        self.assertEqual(
            build_pii_safe_page_url(url='/groups/test/'),
            '/groups/test/')


class BuildPIISafePageURLForRequestTests(TestCase):
    """Unit tests for build_pii_safe_page_url_for_request."""

    def test_with_pii_keyword_in_path(self):
        """Testing build_pii_safe_page_url_for_request with PII
        keyword-matching in URL path
        """
        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/users/test/', {
                    'user': 'test',
                })),
            '/users/<REDACTED>/')

        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/users/test/', {
                    'username': 'test',
                })),
            '/users/<REDACTED>/')

        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/users/test/', {
                    'email': 'test',
                })),
            '/users/<REDACTED>/')

    def test_with_pii_keyword_in_querystring(self):
        """Testing build_pii_safe_page_url_for_request with PII
        keyword-matching in URL query string
        """
        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/?user=test')),
            '/?user=<REDACTED>')

        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/?username=test')),
            '/?username=<REDACTED>')

        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/?email=test')),
            '/?email=<REDACTED>')

    def test_with_pii_email_in_path(self):
        """Testing build_pii_safe_page_url_for_request with email-like value
        in URL path
        """
        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/test/a@b.com/', {
                    'foo': 'a@b.com',
                })),
            '/test/<REDACTED>/')

    def test_with_pii_email_in_querystring(self):
        """Testing build_pii_safe_page_url_for_request with email-like value
        in URL query string
        """
        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/?foo=a@b.com')),
            '/?foo=<REDACTED>')

    def test_with_no_pii(self):
        """Testing build_pii_safe_page_url_for_request with no PII"""
        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/groups/test/', {
                    'groupname': 'test',
                })),
            '/groups/test/')

    def test_with_no_keywords(self):
        """Testing build_pii_safe_page_url_for_request with no keywords in URL
        pattern
        """
        self.assertEqual(
            build_pii_safe_page_url_for_request(
                self._build_request('/groups/test/')),
            '/groups/test/')

    def test_with_no_resolver_match(self):
        """Testing build_pii_safe_page_url_for_request with no resolver match
        """
        self.assertEqual(
            build_pii_safe_page_url_for_request(
                RequestFactory().get('/groups/test/')),
            '/groups/test/')

    def _build_request(self, path, kwargs={}):
        request = RequestFactory().get(path)
        request.resolver_match = ResolverMatch(func=lambda: None,
                                               args=(),
                                               kwargs=kwargs)

        return request



#
# NOTE: This file is deprecated, but due to being referenced in evolutions
#       and other places, we do not warn in this file. We also must keep
#       this file around for JSONField, CounterField, Base64Field, and
#       ModificationTimestampField indefinitely.
#

from djblets.db.fields import (Base64DecodedValue, Base64Field,
                               Base64FieldCreator, CounterField, JSONField,
                               ModificationTimestampField)
from djblets.db.validators import validate_json


__all__ = [
    'Base64DecodedValue',
    'Base64Field',
    'Base64FieldCreator',
    'CounterField',
    'JSONField',
    'ModificationTimestampField',
    'validate_json',
]



"""Unit tests for djblets.integrations.forms.IntegrationConfigForm."""

from django.test.client import RequestFactory

from djblets.integrations.forms import IntegrationConfigForm
from djblets.integrations.integration import Integration
from djblets.integrations.manager import IntegrationManager
from djblets.integrations.tests.models import IntegrationConfig
from djblets.integrations.tests.testcases import IntegrationsTestCase


class DummyIntegration(Integration):
    def initialize(self):
        pass

    def shutdown(self):
        pass

    def _on_callback(self, **kwargs):
        pass


class IntegrationConfigFormTests(IntegrationsTestCase):
    """Unit tests for djblets.integrations.forms.IntegrationConfigForm."""

    def setUp(self):
        super(IntegrationConfigFormTests, self).setUp()

        self.manager = IntegrationManager(IntegrationConfig)
        self.integration = \
            self.manager.register_integration_class(DummyIntegration)
        request_factory = RequestFactory()
        self.request = request_factory.get('/getSomeForm/')

        self.form = IntegrationConfigForm(
            integration=self.integration,
            request=self.request,
            instance=None,
            data={
                'enabled': True,
                'name': 'FakeName',
            })
        self.form.full_clean()
        self.form.save()

    def test_config(self):
        """Testing IntegrationConfigForm.config returns the config instance"""
        self.assertIsInstance(self.form.config, IntegrationConfig)

    def test_get_key_value_model_field(self):
        """Testing IntegrationConfigForm.get_key_value() returns values from
        form data"""
        self.assertEqual(self.form.get_key_value('name'), 'FakeName')
        self.assertTrue(self.form.get_key_value('enabled'))

    def test_set_key_value_model_field(self):
        """Testing IntegrationConfigForm.set_key_value() overrides form
        data"""
        name_field = 'name'
        expected_value = 'NewName'
        self.form.set_key_value(name_field, expected_value)
        self.assertEqual(self.form.get_key_value(name_field), expected_value)

    def test_set_key_value_non_model_field(self):
        """Testing IntegrationConfigForm.set_key_value() sets custom non-field
        data"""
        test_field = 'new_non_model_field'
        self.form.set_key_value(test_field, True)
        self.assertTrue(self.form.get_key_value(test_field))



"""Unit tests for djblets.forms.widgets.ListEditWidget."""

from django import forms
from django.utils.datastructures import MultiValueDict

from djblets.deprecation import RemovedInDjblets40Warning
from djblets.forms.fields import ListEditField
from djblets.forms.widgets import ListEditWidget
from djblets.testing.testcases import TestCase


class ListEditWidgetTests(TestCase):
    """Unit tests for djblets.forms.widgets.ListEditWidget."""

    def test_render(self):
        """Testing ListEditWidget.render with a string of separated values"""
        field = forms.CharField(widget=ListEditWidget())
        rendered = field.widget.render(
            name='my_field',
            value=' foo,  bar , baz ',
            attrs={
                'id': 'id_my_field',
                'class': 'my-value-class',
            })

        self.assertIn(
            '<div class="djblets-c-list-edit-widget list-edit-widget"'
            ' id="id_my_field_container">',
            rendered)
        self.assertIn('<input type="text"'
                      ' name="my_field_value[0]"'
                      ' value="foo"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input type="text"'
                      ' name="my_field_value[1]"'
                      ' value="bar"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input type="text"'
                      ' name="my_field_value[2]"'
                      ' value="baz"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)

    def test_render_with_custom_separator(self):
        """Testing ListEditWidget.render with a string of separated values
           using a custom separator"""
        field = forms.CharField(widget=ListEditWidget(sep=';'))
        rendered = field.widget.render(
            name='my_field',
            value=' foo;  bar ; baz ',
            attrs={
                'id': 'id_my_field',
                'class': 'my-value-class',
            })

        self.assertIn(
            '<div class="djblets-c-list-edit-widget list-edit-widget"'
            ' id="id_my_field_container">',
            rendered)
        self.assertIn('<input type="text"'
                      ' name="my_field_value[0]"'
                      ' value="foo"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input type="text"'
                      ' name="my_field_value[1]"'
                      ' value="bar"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input type="text"'
                      ' name="my_field_value[2]"'
                      ' value="baz"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)

    def test_render_with_custom_value_widget(self):
        """Testing ListEditWidget.render with using a custom value widget"""
        field = ListEditField(
            widget=ListEditWidget(value_widget=forms.widgets.EmailInput))
        rendered = field.widget.render(
            name='my_field',
            value=['a@test.com', 'b@test.com'],
            attrs={
                'id': 'id_my_field',
                'class': 'my-value-class',
            })

        self.assertIn(
            '<div class="djblets-c-list-edit-widget list-edit-widget"'
            ' id="id_my_field_container">',
            rendered)
        self.assertIn('<input type="email"'
                      ' name="my_field_value[0]"'
                      ' value="a@test.com"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)
        self.assertIn('<input type="email"'
                      ' name="my_field_value[1]"'
                      ' value="b@test.com"'
                      ' class="my-value-class'
                      ' djblets-c-list-edit-widget__input">',
                      rendered)

    def test_value_from_datadict(self):
        """Testing ListEditWidget.value_from_datadict"""
        field = ListEditField(widget=ListEditWidget())
        data = MultiValueDict('')
        data.update({
            'my_field_value[0]': 'a@test.com',
            'my_field_value[1]': 'b@test.com',
            'my_field_num_rows': '2',
            'my_field_use_legacy_behavior': 'false'
        })

        self.assertEqual(
            field.widget.value_from_datadict(data, MultiValueDict(''),
                                             'my_field'),
            ['a@test.com', 'b@test.com'])

    def test_value_from_datadict_using_field_name(self):
        """Testing ListEditWidget.value_from_datadict when only the field
        name is given in the data dict"""
        field = ListEditField(widget=ListEditWidget())
        data = MultiValueDict('')
        data.update({
            'my_field': ['a@test.com', 'b@test.com']
        })

        self.assertEqual(
            field.widget.value_from_datadict(data, MultiValueDict(''),
                                             'my_field'),
            ['a@test.com', 'b@test.com'])

    def test_value_from_datadict_using_field_name_and_legacy_behavior(self):
        """Testing ListEditWidget.value_from_datadict when only the field
        name is given in the data dict and using legacy behavior"""
        field = ListEditField(widget=ListEditWidget())
        data = MultiValueDict('')
        data.update({
            'my_field': 'a@test.com,b@test.com',
            'my_field_use_legacy_behavior': 'true'
        })

        self.assertEqual(
            field.widget.value_from_datadict(data, MultiValueDict(''),
                                             'my_field'),
            'a@test.com,b@test.com')
        self.assertWarns(RemovedInDjblets40Warning,
                         ('Passing a string to this widget is deprecated '
                          'and will be removed in 4.0. Pass a list instead.'))

    def test_value_from_datadict_using_legacy_behavior(self):
        """Testing ListEditWidget.value_from_datadict using legacy behavior"""
        field = ListEditField(widget=ListEditWidget())
        data = MultiValueDict('')
        data.update({
            'my_field_value[0]': 'a@test.com',
            'my_field_value[1]': 'b@test.com',
            'my_field_num_rows': '2',
            'my_field_use_legacy_behavior': 'true'
        })

        self.assertEqual(
            field.widget.value_from_datadict(data, MultiValueDict(''),
                                             'my_field'),
            'a@test.com,b@test.com')
        self.assertWarns(RemovedInDjblets40Warning,
                         ('Passing a string to this widget is deprecated '
                          'and will be removed in 4.0. Pass a list instead.'))

    def test_get_context(self):
        """Testing ListEditWidget.get_context"""
        field = ListEditField(widget=ListEditWidget())
        result = field.widget.get_context(
            'my_field',
            ['foo', 'bar'],
            {
                'id': 'id_my_field',
                'class': 'my-value-class'
            })

        self.assertEqual(
            set(result),
            {
                'name', 'id', 'remove_text', 'rendered_rows',
                'rendered_empty_row'
            })
        self.assertEqual(result['name'], 'my_field')
        self.assertEqual(result['id'], 'id_my_field')

        rendered_rows = result['rendered_rows']
        self.assertEqual(len(rendered_rows), 2)
        self.assertHTMLEqual(
            rendered_rows[0],
            '<input type="text"'
            ' name="my_field_value[0]"'
            ' value="foo"'
            ' class="my-value-class'
            ' djblets-c-list-edit-widget__input" />')
        self.assertHTMLEqual(
            rendered_rows[1],
            '<input type="text"'
            ' name="my_field_value[1]"'
            ' value="bar"'
            ' class="my-value-class'
            ' djblets-c-list-edit-widget__input" />')
        self.assertHTMLEqual(
            result['rendered_empty_row'],
            '<input type="text"'
            ' name="my_field_value[0]"'
            ' class="my-value-class'
            ' djblets-c-list-edit-widget__input" />')



"""Tests for djblets.gravatars."""

from hashlib import md5

from django.contrib.auth.models import User
from django.template import Context, Template

from djblets.gravatars import get_gravatar_url, get_gravatar_url_for_email
from djblets.testing.testcases import TagTest, TestCase


class GravatarTests(TestCase):
    """Tests for gravatar functions."""

    _URL_BASE = 'https://secure.gravatar.com/avatar/'

    def test_get_gravatar_url_for_email_uppercase(self):
        """Testing get_gravatar_url_for_email with uppercase characters"""
        self.assertEqual(get_gravatar_url_for_email(email='FOO@EXAMPLE.COM'),
                         get_gravatar_url_for_email(email='foo@example.com'))
        self.assertEqual(
            get_gravatar_url_for_email(email='foo@example.com'),
            '%s%s' % (self._URL_BASE, md5(b'foo@example.com').hexdigest()))

    def test_get_gravatar_url_for_email_whitespace(self):
        """Testing get_gravatar_url_for_email with whitespace characters"""
        self.assertEqual(get_gravatar_url_for_email(email=' foo@example.com '),
                         get_gravatar_url_for_email(email='foo@example.com'))

    def test_get_gravatar_url_for_email_unicode(self):
        """Testing get_gravatar_url_for_email with unicode characters"""
        raw_email = 'こんにちは@example.com'
        encoded_email = raw_email.encode('utf-8')

        self.assertEqual(get_gravatar_url_for_email(email=raw_email),
                         get_gravatar_url_for_email(email=encoded_email))
        self.assertEqual(
            get_gravatar_url_for_email(email=raw_email),
            '%s%s' % (self._URL_BASE, md5(encoded_email).hexdigest()))

    def test_get_gravatar_url_for_email_query_order(self):
        """Testing get_gravatar_url_for_email query string ordering"""
        email = 'foo@example.com'

        with self.settings(GRAVATAR_RATING='G',
                           GRAVATAR_DEFAULT='my-default'):
            self.assertEqual(
                get_gravatar_url_for_email(email='foo@example.com', size=128),
                '%s%s?s=128&r=G&d=my-default'
                % (self._URL_BASE, md5(email.encode('utf-8')).hexdigest()))

    def test_get_gravatar_url_uppercase(self):
        """Testing get_gravatar_url with uppercase characters"""
        user = User.objects.create_user(username='user',
                                        email='FOO@EXAMPLE.COM')
        self.assertEqual(get_gravatar_url(user=user),
                         get_gravatar_url_for_email(email='foo@example.com'))

    def test_get_gravatar_url_whitespace(self):
        """Testing get_gravatar_url with whitespace characters"""
        user = User.objects.create_user(username='user',
                                        email=' foo@example.com ')
        self.assertEqual(get_gravatar_url(user=user),
                         get_gravatar_url_for_email(email='foo@example.com'))

    def test_get_gravatar_url_for_unicode(self):
        """Testing get_gravatar_url with unicode characters"""
        raw_email = 'こんにちは@example.com'
        encoded_email = raw_email.encode('utf-8')
        user = User.objects.create_user(username='user', email=raw_email)
        self.assertEqual(get_gravatar_url(user=user),
                         get_gravatar_url_for_email(email=encoded_email))


class TagTests(TagTest):
    """Unit tests for gravatars template tags."""

    def test_gravatar_xss(self):
        """Testing {% gravatar %} doesn't allow XSS injection"""
        user = User(username='test',
                    first_name='"><script>alert(1);</script><"',
                    email='test@example.com')

        t = Template('{% load gravatars %}'
                     '{% gravatar user 32 %}')

        self.assertEqual(
            t.render(Context({
                'user': user,
            })),
            '<img src="https://secure.gravatar.com/avatar/'
            '55502f40dc8b7c769880b10874abc9d0?s=32" width="32" height="32" '
            'alt="&quot;&gt;&lt;script&gt;alert(1);&lt;/script&gt;&lt;&quot;" '
            'class="gravatar"/>')

    def test_gravatar_url_tag(self):
        """Testing gravatar_url template tag"""
        t = Template('{% load gravatars %}'
                     '{% gravatar_url "test@example.com" 32 %}')

        self.assertEqual(
            t.render(Context({})),
            'https://secure.gravatar.com/avatar/'
            '55502f40dc8b7c769880b10874abc9d0?s=32')



import calendar
import datetime

from django import template


register = template.Library()


@register.filter
def feeddate(datetuple):
    """
    A filter that converts the date tuple provided from feedparser into
    a datetime object.
    """
    return datetime.datetime.utcfromtimestamp(calendar.timegm(datetuple))



import inspect
import pickle
import zlib

import kgb
from django.contrib.sites.models import Site
from django.core.cache import cache
from django.core.cache.backends.base import memcache_key_warnings
from django.test.utils import override_settings

from djblets.cache.backend import (CACHE_CHUNK_SIZE,
                                   MAX_KEY_SIZE,
                                   cache_memoize,
                                   cache_memoize_iter,
                                   make_cache_key,
                                   _get_default_encryption_key)
from djblets.secrets.crypto import AES_BLOCK_SIZE, aes_decrypt, aes_encrypt
from djblets.testing.testcases import TestCase


class BaseCacheTestCase(kgb.SpyAgency, TestCase):
    """Base class for cache-related unit tests."""

    CUSTOM_ENCRYPTION_KEY = b'0123456789abcdef'

    def tearDown(self):
        super().tearDown()

        cache.clear()

    def build_test_chunk_data(self, num_chunks, extra_len=0,
                              use_compression=False, use_encryption=False,
                              encryption_key=None):
        """Build enough test data to fill up the specified number of chunks.

        This takes into account the size of the pickle data, and will
        get us to exactly the specified number of chunks of data in the cache.

        Args:
            num_chunks (int):
                The number of chunks to build.

            extra_len (int, optional):
                Extra length to add to the data size.

            use_compression (bool, optional):
                Whether to compress the payload.

            use_encryption (bool, optional):
                Whether to generate a payload suitable for encryption.

            encryption_key (bytes, optional):
                An optional non-default encryption key.

        Returns:
            tuple:
            A 2-tuple containing:

            1. The raw generated data.
            2. The resulting chunk data to store.
        """
        data_len = CACHE_CHUNK_SIZE * num_chunks - 3 * num_chunks + extra_len

        if use_encryption:
            data_len -= AES_BLOCK_SIZE

        data = 'x' * data_len

        chunk_data = pickle.dumps(data, protocol=0)
        self.assertTrue(chunk_data.startswith(b'Vxxxxxxxx'))

        if use_compression:
            chunk_data = zlib.compress(chunk_data)

        if use_encryption:
            chunk_data = aes_encrypt(
                chunk_data,
                key=encryption_key or _get_default_encryption_key())

        if not use_compression:
            self.assertEqual(len(chunk_data),
                             CACHE_CHUNK_SIZE * num_chunks + extra_len)

        return data, chunk_data


class CacheMemoizeTests(BaseCacheTestCase):
    """Unit tests for cache_memoize."""

    def test_with_string(self):
        """Testing cache_memoize with string"""
        cache_key = 'abc123'
        test_str = 'Test 123'

        def cache_func():
            return test_str

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key, cache_func)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_unicode_data(self):
        """Testing cache_memoize with Unicode data"""
        cache_key = 'abc123'
        test_str = '🙃' * 5

        def cache_func():
            return test_str

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key, cache_func)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_non_sequence(self):
        """Testing cache_memoize with non-sequence type data"""
        cache_key = 'abc123'
        data = True

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key, cache_func)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_large_files_uncompressed(self):
        """Testing cache_memoize with large files without compression"""
        cache_key = 'abc123'

        # This takes into account the size of the pickle data, and will
        # get us to exactly 2 chunks of data in cache.
        data, pickled_data = self.build_test_chunk_data(num_chunks=2)

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=False)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

        cache_key_0 = make_cache_key('%s-0' % cache_key)
        cache_key_1 = make_cache_key('%s-1' % cache_key)

        self.assertIn(make_cache_key(cache_key), cache)
        self.assertIn(cache_key_0, cache)
        self.assertIn(cache_key_1, cache)
        self.assertNotIn(make_cache_key('%s-2' % cache_key), cache)

        # Verify the contents of the stored data.
        stored_data = b''.join(cache.get(cache_key_0) + cache.get(cache_key_1))
        self.assertEqual(stored_data, pickled_data)

        # Try fetching the data we stored.
        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=False)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_large_files_uncompressed_off_by_one(self):
        """Testing cache_memoize with large files without compression and
        one byte larger than an even chunk size
        """
        cache_key = 'abc123'

        # This takes into account the size of the pickle data, and will
        # get us to just barely 3 chunks of data in cache.
        data, pickled_data = self.build_test_chunk_data(num_chunks=2,
                                                        extra_len=1)

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=False)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

        cache_key_0 = make_cache_key('%s-0' % cache_key)
        cache_key_1 = make_cache_key('%s-1' % cache_key)
        cache_key_2 = make_cache_key('%s-2' % cache_key)

        self.assertIn(make_cache_key(cache_key), cache)
        self.assertIn(cache_key_0, cache)
        self.assertIn(cache_key_1, cache)
        self.assertIn(cache_key_2, cache)
        self.assertNotIn(make_cache_key('%s-3' % cache_key), cache)

        # Verify the contents of the stored data.
        stored_data = b''.join(cache.get(cache_key_0) +
                               cache.get(cache_key_1) +
                               cache.get(cache_key_2))
        self.assertEqual(stored_data, pickled_data)

        # Try fetching the data we stored.
        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=False)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_large_files_compressed(self):
        """Testing cache_memoize with large files with compression"""
        cache_key = 'abc123'

        data, pickled_data = self.build_test_chunk_data(num_chunks=2)

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=True)
        self.assertSpyCallCount(cache_func, 1)

        cache_key_0 = make_cache_key('%s-0' % cache_key)

        self.assertIn(make_cache_key(cache_key), cache)
        self.assertIn(cache_key_0, cache)
        self.assertNotIn(make_cache_key('%s-1' % cache_key), cache)
        self.assertNotIn(make_cache_key('%s-2' % cache_key), cache)

        # Verify the contents of the stored data.
        stored_data = cache.get(cache_key_0)[0]
        self.assertEqual(stored_data, zlib.compress(pickled_data))

        # Try fetching the data we stored.
        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_large_files_load_uncompressed(self):
        """Testing cache_memoize with large files without compression and
        loading data
        """
        cache_key = 'abc123'

        # This takes into account the size of the pickle data, and will
        # get us to exactly 2 chunks of data in cache.
        data, pickled_data = self.build_test_chunk_data(num_chunks=2)

        cache.set(make_cache_key(cache_key), '2')
        cache.set(make_cache_key('%s-0' % cache_key),
                  [pickled_data[:CACHE_CHUNK_SIZE]])
        cache.set(make_cache_key('%s-1' % cache_key),
                  [pickled_data[CACHE_CHUNK_SIZE:]])

        def cache_func():
            return ''

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=False)
        self.assertEqual(result, data)
        self.assertSpyNotCalled(cache_func)

    def test_with_large_files_load_compressed(self):
        """Testing cache_memoize with large files with compression and
        loading cached data
        """
        cache_key = 'abc123'

        data, stored_data = self.build_test_chunk_data(
            num_chunks=2,
            use_compression=True)
        self.assertLess(len(stored_data), CACHE_CHUNK_SIZE)

        cache.set(make_cache_key(cache_key), '1')
        cache.set(make_cache_key('%s-0' % cache_key), [stored_data])

        def cache_func():
            return ''

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=True)
        self.assertEqual(result, data)
        self.assertSpyNotCalled(cache_func)

    def test_with_large_files_missing_chunk(self):
        """Testing cache_memoize with loading large files with missing chunks
        """
        cache_key = 'abc123'

        # This takes into account the size of the pickle data, and will
        # get us to exactly 2 chunks of data in cache.
        data, pickled_data = self.build_test_chunk_data(num_chunks=2)

        cache.set(make_cache_key(cache_key), '2')
        cache.set(make_cache_key('%s-0' % cache_key),
                  [pickled_data[:CACHE_CHUNK_SIZE]])

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func, large_data=True,
                               compress_large_data=False)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption(self):
        """Testing cache_memoize with use_encryption=True"""
        cache_key = 'abc123'
        test_str = 'Test 123'

        def cache_func():
            return test_str

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               use_encryption=True)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        encrypted_key = \
            '0a3195ea908f9bd178ab3c67a257ea4e7954e4b67ac240d1c86953953cde3a0b'
        self.assertIn(encrypted_key, cache)

        cache_value = cache.get(encrypted_key)
        self.assertNotEqual(cache_value, test_str)
        self.assertEqual(pickle.loads(aes_decrypt(cache_value)),
                         test_str)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key,
                               cache_func,
                               use_encryption=True)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_custom_encryption_key(self):
        """Testing cache_memoize with use_encryption=True and custom
        encryption_key
        """
        cache_key = 'abc123'
        test_str = 'Test 123'

        def cache_func():
            return test_str

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               use_encryption=True,
                               encryption_key=self.CUSTOM_ENCRYPTION_KEY)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        encrypted_key = \
            '10c4714f5c6c49045ed25c8da7d04a4be2254fa275fc3b63bdeebe887aee52f9'
        self.assertIn(encrypted_key, cache)

        cache_value = cache.get(encrypted_key)
        self.assertNotEqual(cache_value, test_str)
        self.assertEqual(
            pickle.loads(aes_decrypt(cache_value,
                                     key=self.CUSTOM_ENCRYPTION_KEY)),
            test_str)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key,
                               cache_func,
                               use_encryption=True,
                               encryption_key=self.CUSTOM_ENCRYPTION_KEY)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_key_changed(self):
        """Testing cache_memoize with use_encryption=True and encryption
        key changed
        """
        cache_key = 'abc123'
        test_str = 'Test 123'

        def cache_func():
            return test_str

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               use_encryption=True)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        encrypted_key = \
            '0a3195ea908f9bd178ab3c67a257ea4e7954e4b67ac240d1c86953953cde3a0b'
        self.assertIn(encrypted_key, cache)

        cache_value = cache.get(encrypted_key)
        self.assertNotEqual(cache_value, test_str)
        self.assertEqual(
            pickle.loads(aes_decrypt(cache_value)),
            test_str)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key,
                               cache_func,
                               use_encryption=True,
                               encryption_key=self.CUSTOM_ENCRYPTION_KEY)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 2)

        encrypted_key = \
            '10c4714f5c6c49045ed25c8da7d04a4be2254fa275fc3b63bdeebe887aee52f9'
        self.assertIn(encrypted_key, cache)

        cache_value = cache.get(encrypted_key)
        self.assertNotEqual(cache_value, test_str)
        self.assertEqual(
            pickle.loads(aes_decrypt(cache_value,
                                     key=self.CUSTOM_ENCRYPTION_KEY)),
            test_str)

    @override_settings(DJBLETS_CACHE_FORCE_ENCRYPTION=True)
    def test_with_force_use_encryption(self):
        """Testing cache_memoize with
        settings.DJBLETS_CACHE_FORCE_ENCRYPTION=True
        """
        cache_key = 'abc123'
        test_str = 'Test 123'

        def cache_func():
            return test_str

        self.spy_on(cache_func)

        result = cache_memoize(cache_key, cache_func)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        encrypted_key = \
            '0a3195ea908f9bd178ab3c67a257ea4e7954e4b67ac240d1c86953953cde3a0b'
        self.assertIn(encrypted_key, cache)

        cache_value = cache.get(encrypted_key)
        self.assertNotEqual(cache_value, test_str)
        self.assertEqual(pickle.loads(aes_decrypt(cache_value)),
                         test_str)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key, cache_func)
        self.assertEqual(result, test_str)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_non_string_data(self):
        """Testing cache_memoize with use_encryption=True and non-string
        data
        """
        cache_key = 'abc123'
        test_data = {
            'item1': True,
            'item2': {1, 2, 3},
            'item3': None,
        }

        def cache_func():
            return test_data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               use_encryption=True)
        self.assertEqual(result, test_data)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        encrypted_key = \
            '0a3195ea908f9bd178ab3c67a257ea4e7954e4b67ac240d1c86953953cde3a0b'
        self.assertIn(encrypted_key, cache)

        cache_value = cache.get(encrypted_key)
        self.assertNotEqual(cache_value, test_data)
        self.assertEqual(pickle.loads(aes_decrypt(cache_value)),
                         test_data)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key, cache_func,
                               use_encryption=True)
        self.assertEqual(result, test_data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_large_files_uncompressed(self):
        """Testing cache_memoize with use_encryption=True and large files
        without compression
        """
        cache_key = 'abc123'

        # This takes into account the size of the encrypted pickle data, and
        # will get us to exactly 2 chunks of data in cache.
        data, encrypted_data = self.build_test_chunk_data(
            num_chunks=2,
            use_encryption=True)

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=False,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        cache_key_main = make_cache_key(cache_key,
                                        use_encryption=True)
        cache_key_0 = make_cache_key('%s-0' % cache_key,
                                     use_encryption=True)
        cache_key_1 = make_cache_key('%s-1' % cache_key,
                                     use_encryption=True)
        cache_key_2 = make_cache_key('%s-2' % cache_key,
                                     use_encryption=True)

        self.assertIn(cache_key_main, cache)
        self.assertIn(cache_key_0, cache)
        self.assertIn(cache_key_1, cache)
        self.assertNotIn(cache_key_2, cache)

        # Verify the contents of the stored data.
        stored_data = b''.join(cache.get(cache_key_0) +
                               cache.get(cache_key_1))
        self.assertNotEqual(stored_data, data)
        self.assertEqual(pickle.loads(aes_decrypt(stored_data)),
                         data)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=False,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_large_files_uncompressed_off_by_one(self):
        """Testing cache_memoize with use_encryption=True and large files
        without compression and one byte larger than an even chunk size
        """
        cache_key = 'abc123'

        # This takes into account the size of the encrypted pickle data, and
        # will get us to just barely 3 chunks of data in cache.
        data, encrypted_data = self.build_test_chunk_data(
            num_chunks=2,
            extra_len=1,
            use_encryption=True)

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=False,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        cache_key_main = make_cache_key(cache_key,
                                        use_encryption=True)
        cache_key_0 = make_cache_key('%s-0' % cache_key,
                                     use_encryption=True)
        cache_key_1 = make_cache_key('%s-1' % cache_key,
                                     use_encryption=True)
        cache_key_2 = make_cache_key('%s-2' % cache_key,
                                     use_encryption=True)
        cache_key_3 = make_cache_key('%s-3' % cache_key,
                                     use_encryption=True)

        self.assertIn(cache_key_main, cache)
        self.assertIn(cache_key_0, cache)
        self.assertIn(cache_key_1, cache)
        self.assertIn(cache_key_2, cache)
        self.assertNotIn(cache_key_3, cache)

        # Verify the contents of the stored data.
        stored_data = b''.join(cache.get(cache_key_0) +
                               cache.get(cache_key_1) +
                               cache.get(cache_key_2))
        self.assertNotEqual(stored_data, data)
        self.assertEqual(pickle.loads(aes_decrypt(stored_data)),
                         data)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=False,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_large_files_compressed(self):
        """Testing cache_memoize with use_encryption=True and large files
        with compression
        """
        cache_key = 'abc123'

        # This takes into account the size of the encrypted pickle data, and
        # will get us to just barely 3 chunks of data in cache.
        data, encrypted_data = self.build_test_chunk_data(
            num_chunks=2,
            use_encryption=True)

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=True,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        cache_key_main = make_cache_key(cache_key,
                                        use_encryption=True)
        cache_key_0 = make_cache_key('%s-0' % cache_key,
                                     use_encryption=True)
        cache_key_1 = make_cache_key('%s-1' % cache_key,
                                     use_encryption=True)
        cache_key_2 = make_cache_key('%s-2' % cache_key,
                                     use_encryption=True)

        self.assertIn(cache_key_main, cache)
        self.assertIn(cache_key_0, cache)
        self.assertNotIn(cache_key_1, cache)
        self.assertNotIn(cache_key_2, cache)

        # Verify the contents of the stored data.
        stored_data = cache.get(cache_key_0)[0]
        self.assertNotEqual(stored_data, data)
        self.assertEqual(
            pickle.loads(zlib.decompress(aes_decrypt(stored_data))),
            data)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=True,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_large_files_load_uncompressed(self):
        """Testing cache_memoize with use_encryption=True and large files
        without compression and loading data
        """
        cache_key = 'abc123'

        # This takes into account the size of the encrypted pickle data, and
        # will get us to just barely 3 chunks of data in cache.
        data, encrypted_data = self.build_test_chunk_data(
            num_chunks=2,
            use_encryption=True)

        cache_key_main = make_cache_key(cache_key,
                                        use_encryption=True)
        cache_key_0 = make_cache_key('%s-0' % cache_key,
                                     use_encryption=True)
        cache_key_1 = make_cache_key('%s-1' % cache_key,
                                     use_encryption=True)

        cache.set(cache_key_main, aes_encrypt(pickle.dumps('2', protocol=0)))
        cache.set(cache_key_0, [encrypted_data[:CACHE_CHUNK_SIZE]])
        cache.set(cache_key_1, [encrypted_data[CACHE_CHUNK_SIZE:]])

        def cache_func():
            return ''

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=False,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyNotCalled(cache_func)

    def test_with_use_encryption_and_large_files_load_compressed(self):
        """Testing cache_memoize with use_encryption=True and large files
        with compression and loading data
        """
        cache_key = 'abc123'

        # This takes into account the size of the encrypted pickle data, and
        # will get us to just barely 3 chunks of data in cache.
        data, encrypted_data = self.build_test_chunk_data(
            num_chunks=2,
            use_compression=True,
            use_encryption=True)
        self.assertLess(len(encrypted_data), CACHE_CHUNK_SIZE)

        cache_key_main = make_cache_key(cache_key,
                                        use_encryption=True)
        cache_key_0 = make_cache_key('%s-0' % cache_key,
                                     use_encryption=True)

        cache.set(cache_key_main, aes_encrypt(pickle.dumps('1', protocol=0)))
        cache.set(cache_key_0, [encrypted_data])

        def cache_func():
            return ''

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=True,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyNotCalled(cache_func)

    def test_with_use_encryption_and_large_files_custom_key(self):
        """Testing cache_memoize with use_encryption=True and large files
        with custom encryption key
        """
        cache_key = 'abc123'
        encryption_key = self.CUSTOM_ENCRYPTION_KEY

        # This takes into account the size of the encrypted pickle data, and
        # will get us to just barely 3 chunks of data in cache.
        data, encrypted_data = self.build_test_chunk_data(
            num_chunks=2,
            use_encryption=True,
            encryption_key=encryption_key)

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=True,
                               use_encryption=True,
                               encryption_key=encryption_key)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

        # Check what we put into cache.
        cache_key_main = make_cache_key(cache_key,
                                        use_encryption=True,
                                        encryption_key=encryption_key)
        cache_key_0 = make_cache_key('%s-0' % cache_key,
                                     use_encryption=True,
                                     encryption_key=encryption_key)
        cache_key_1 = make_cache_key('%s-1' % cache_key,
                                     use_encryption=True,
                                     encryption_key=encryption_key)
        cache_key_2 = make_cache_key('%s-2' % cache_key,
                                     use_encryption=True,
                                     encryption_key=encryption_key)

        self.assertIn(cache_key_main, cache)
        self.assertIn(cache_key_0, cache)
        self.assertNotIn(cache_key_1, cache)
        self.assertNotIn(cache_key_2, cache)

        # Verify the contents of the stored data.
        stored_data = cache.get(cache_key_0)[0]
        self.assertNotEqual(stored_data, data)
        self.assertEqual(
            pickle.loads(zlib.decompress(
                aes_decrypt(stored_data, key=encryption_key))),
            data)

        # Call a second time. We should only call cache_func once.
        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=True,
                               use_encryption=True,
                               encryption_key=encryption_key)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_large_files_missing_chunk(self):
        """Testing cache_memoize with use_encryption=True and large files
        with missing chunks
        """
        cache_key = 'abc123'

        # This takes into account the size of the encrypted pickle data, and
        # will get us to just barely 3 chunks of data in cache.
        data, encrypted_data = self.build_test_chunk_data(
            num_chunks=2,
            use_encryption=True)

        cache_key_main = make_cache_key(cache_key,
                                        use_encryption=True)
        cache_key_0 = make_cache_key('%s-0' % cache_key,
                                     use_encryption=True)

        cache.set(cache_key_main, aes_encrypt(pickle.dumps('2', protocol=0)))
        cache.set(cache_key_0, [encrypted_data[:CACHE_CHUNK_SIZE]])

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=False,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)

    def test_with_use_encryption_and_large_files_bad_chunk(self):
        """Testing cache_memoize with use_encryption=True and large files
        with chunk failing decryption
        """
        cache_key = 'abc123'

        # This takes into account the size of the encrypted pickle data, and
        # will get us to just barely 3 chunks of data in cache.
        data, encrypted_data = self.build_test_chunk_data(
            num_chunks=2,
            use_encryption=True,
            encryption_key=self.CUSTOM_ENCRYPTION_KEY)

        cache_key_main = make_cache_key(cache_key,
                                        use_encryption=True)
        cache_key_0 = make_cache_key('%s-0' % cache_key,
                                     use_encryption=True)

        cache.set(cache_key_main, aes_encrypt(pickle.dumps('2', protocol=0)))
        cache.set(cache_key_0, [encrypted_data[:CACHE_CHUNK_SIZE]])

        def cache_func():
            return data

        self.spy_on(cache_func)

        result = cache_memoize(cache_key,
                               cache_func,
                               large_data=True,
                               compress_large_data=False,
                               use_encryption=True)
        self.assertEqual(result, data)
        self.assertSpyCallCount(cache_func, 1)


class CacheMemoizeIterTests(BaseCacheTestCase):
    """Unit tests for cache_memoize_iter."""

    def test_without_compression(self):
        """Testing cache_memoize_iter without compression"""
        cache_key = 'abc123'
        data_yielded = []

        data1, pickled_data_1 = self.build_test_chunk_data(num_chunks=2)
        data2, pickled_data_2 = self.build_test_chunk_data(num_chunks=2)

        def cache_func():
            data_yielded.append('data1')
            yield data1

            data_yielded.append('data2')
            yield data2

        self.spy_on(cache_func)

        result = cache_memoize_iter(cache_key, cache_func,
                                    compress_large_data=False)
        self.assertTrue(inspect.isgenerator(result))
        self.assertEqual(data_yielded, [])

        self.assertEqual(next(result), data1)
        self.assertEqual(data_yielded, ['data1'])

        self.assertEqual(next(result), data2)
        self.assertEqual(data_yielded, ['data1', 'data2'])

        with self.assertRaises(StopIteration):
            next(result)

        self.assertSpyCallCount(cache_func, 1)

        cache_key_main = make_cache_key(cache_key)
        cache_key_0 = make_cache_key('%s-0' % cache_key)
        cache_key_1 = make_cache_key('%s-1' % cache_key)
        cache_key_2 = make_cache_key('%s-2' % cache_key)
        cache_key_3 = make_cache_key('%s-3' % cache_key)

        self.assertIn(cache_key_main, cache)
        self.assertIn(cache_key_0, cache)
        self.assertIn(cache_key_1, cache)
        self.assertIn(cache_key_2, cache)
        self.assertIn(cache_key_3, cache)
        self.assertNotIn(make_cache_key('%s-4' % cache_key), cache)

        # Verify the contents of the stored data.
        stored_data = b''.join(cache.get(cache_key_0) +
                               cache.get(cache_key_1) +
                               cache.get(cache_key_2) +
                               cache.get(cache_key_3))
        self.assertEqual(cache.get(cache_key_main), '4')
        self.assertEqual(stored_data, pickled_data_1 + pickled_data_2)

        # Try fetching the data we stored.
        data_yielded = []

        result = cache_memoize_iter(cache_key, cache_func,
                                    compress_large_data=False)
        self.assertTrue(inspect.isgenerator(result))

        self.assertEqual(next(result), data1)
        self.assertEqual(next(result), data2)

        with self.assertRaises(StopIteration):
            next(result)

        self.assertEqual(data_yielded, [])
        self.assertSpyCallCount(cache_func, 1)

    def test_with_compressed(self):
        """Testing cache_memoize_iter with compression"""
        cache_key = 'abc123'
        data_yielded = []

        # This takes into account the size of the pickle data, and will
        # get us to exactly 2 chunks of data in cache, each.
        data1, pickled_data_1 = self.build_test_chunk_data(num_chunks=2)
        data2, pickled_data_2 = self.build_test_chunk_data(num_chunks=2)

        def cache_func():
            data_yielded.append('data1')
            yield data1

            data_yielded.append('data2')
            yield data2

        self.spy_on(cache_func)

        result = cache_memoize_iter(cache_key, cache_func,
                                    compress_large_data=True)
        self.assertTrue(inspect.isgenerator(result))
        self.assertEqual(data_yielded, [])

        self.assertEqual(next(result), data1)
        self.assertEqual(data_yielded, ['data1'])

        self.assertEqual(next(result), data2)
        self.assertEqual(data_yielded, ['data1', 'data2'])

        with self.assertRaises(StopIteration):
            next(result)

        self.assertSpyCallCount(cache_func, 1)

        cache_key_main = make_cache_key(cache_key)
        cache_key_0 = make_cache_key('%s-0' % cache_key)

        self.assertIn(cache_key_main, cache)
        self.assertIn(cache_key_0, cache)
        self.assertNotIn(make_cache_key('%s-1' % cache_key), cache)

        # Verify the contents of the stored data.
        self.assertEqual(cache.get(cache_key_main), '1')
        self.assertEqual(cache.get(cache_key_0)[0],
                         zlib.compress(pickled_data_1 + pickled_data_2))

        # Try fetching the data we stored.
        data_yielded = []

        result = cache_memoize_iter(cache_key, cache_func,
                                    compress_large_data=True)
        self.assertTrue(inspect.isgenerator(result))
        self.assertEqual(next(result), data1)
        self.assertEqual(next(result), data2)

        with self.assertRaises(StopIteration):
            next(result)

        self.assertEqual(data_yielded, [])
        self.assertSpyCallCount(cache_func, 1)


class MakeCacheKeyTests(BaseCacheTestCase):
    """Unit tests for make_cache_key."""

    def test_default(self):
        """Testing make_cache_key"""
        self._check_key('test123',
                        'example.com:test123')

    @override_settings(SITE_ROOT='/subdir/')
    def test_with_site_root(self):
        """Testing make_cache_key with SITE_ROOT"""
        self._check_key('test123',
                        'example.com:/subdir/:test123')

    def test_without_site(self):
        """Testing make_cache_key without Site object."""
        Site.objects.all().delete()
        self._check_key('test123',
                        'test123')

    def test_with_unicode(self):
        """Testing make_cache_key with Unicode characters"""
        self._check_key('test🥳123',
                        'example.com:test🥳123')

    def test_with_invalid_chars(self):
        """Testing make_cache_key with invalid characters"""
        self._check_key('test\0\tkey !"\x7f',
                        r'example.com:test\x00\x09key\x20!"\x7f')

    def test_with_large_keys_max_length(self):
        """Testing make_cache_key with large keys at max length"""
        length = MAX_KEY_SIZE - len('example.com:')
        self._check_key(
            'x' * length,
            'example.com:xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx')

    def test_with_large_keys_overflow(self):
        """Testing make_cache_key with large keys > max length"""
        length = MAX_KEY_SIZE - len('example.com:') + 1
        self._check_key(
            'x' * length,
            'example.com:xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxedf8474a1b91e737'
            'ec28252abe7a8aa5604539b239505e8ccad8890a44198c58')

    def test_with_encryption(self):
        """Testing make_cache_key with use_encryption=True"""
        self._check_key(
            'abc123',
            '0a3195ea908f9bd178ab3c67a257ea4e7954e4b67ac240d1c86953953cde3a0b',
            use_encryption=True)

    def test_with_encryption_and_custom_encryption_key(self):
        """Testing make_cache_key with use_encryption=True and custom
        encryption_key
        """
        self._check_key(
            'abc123',
            '10c4714f5c6c49045ed25c8da7d04a4be2254fa275fc3b63bdeebe887aee52f9',
            use_encryption=True,
            encryption_key=self.CUSTOM_ENCRYPTION_KEY)

    @override_settings(DJBLETS_CACHE_FORCE_ENCRYPTION=True)
    def test_with_force_encryption_setting(self):
        """Testing make_cache_key with
        settings.DJBLETS_CACHE_FORCE_ENCRYPTION=True
        """
        self._check_key(
            'abc123',
            '0a3195ea908f9bd178ab3c67a257ea4e7954e4b67ac240d1c86953953cde3a0b')

    def _check_key(self, key, expected_key, **kwargs):
        """Check the results of a key, and ensure it works in cache.

        Args:
            key (str):
                The key to build a cache key out of.

            expected_key (str):
                The expected full cache key.

            **kwargs (dict):
                Keyword arguments to pass to
                :py:func:`~djblets.cache.backend.make_cache_key`.

        Raises:
            AssertionError:
                The results were not equal.
        """
        full_cache_key = make_cache_key(key, **kwargs)

        self.assertIsInstance(full_cache_key, str)
        self.assertEqual(full_cache_key, expected_key)

        # Generate a second, to be sure nothing changes.
        full_cache_key2 = make_cache_key(key, **kwargs)
        self.assertEqual(full_cache_key, full_cache_key2)

        # Validate the key.
        self.assertEqual(
            list(memcache_key_warnings(full_cache_key)),
            [])

        # Make sure this can be set and retrieved from cache.
        cache.set(full_cache_key, 'test-value')
        self.assertEqual(cache.get(full_cache_key), 'test-value')
